"""This software was initially develoepd by the US Department of the Interior
in 2014. It was modified and updated by a member of the oringal development team with 
support from CPC, Inc, in 2022. Further iterations were completed by JBH Offshore thereafter. 
This software is provide 'as is'. 
The developers give no warranty, expressed or implied, 
as to the accuracy, reliability,or completeness of this software. 
Although this software has been processed successfully on a computer system at CPC, Inc, no warranty
expressed or implied is made regarding the functioning of the software on
another system or for general or scientific purposes, nor shall the act of
distribution constitute any such warranty. This disclaimer applies both to
individual use of the software and aggregate use with other software.
"""
import sys
import os
from urllib.parse import _NetlocResultMixinBase
import arcpy 
import numpy 
import datetime
import math
import pandas as pd
import scipy.signal as signal
import glob
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
from arcpy.sa import *

class GammaReadings(object):
    def __init__(self):
        self.surveyLineData = {}

    def getLineNumbers(self, data):
        """Returns a set of line numbers from the data dictionary
        Arguments: 
        data - a list of dictionaries
        """
        lineNumberList = []
        for record in data:
            lineNumberList.append(record['lineNumber'])
        lineNumberSet = set(lineNumberList)
        return lineNumberSet

    def createSurveyLineData(self, data):
        """Splits the data dictionary up by line number and populates
        the surveyLineData dictionary.
        Arguments: 
        data - a list of dictionaries
        """
        lineNumbers = self.getLineNumbers(data)
        self.surveyLineData = {}
        for lineNumber in lineNumbers:
            self.surveyLineData['{}'.format(lineNumber)] = [record for record in data if record['lineNumber'] == lineNumber]		

    
    def calculateGradient(self, sensorNoise, back, forward):
        """Calculates the change in total field as the difference between the current and
        the mean of the surrounding total field readings. If the change in
        total field are less than the sensor noise it is reassigned the value of 0 and the
        change in total is assumed to be the result of noise. Does not maintain neighborhood
        size at the survey line ends.
        Aruguments: 
        sensorNoise - float - sensor noise in total field 
        back - int - number previous records to include in the mean background total field 
        forward - int - number of following records to include in the mean background total field 
        """		
        for surveyLine in self.surveyLineData:
            for recordIndex in range(len(self.surveyLineData[surveyLine])):
                values = []
                neighborhood = range(recordIndex - back, recordIndex + forward + 1)
                neighborhood = [index for index in neighborhood if index != recordIndex]
                for searchIndex in neighborhood:
                    if searchIndex >= 0 and searchIndex < len(self.surveyLineData[surveyLine]):
                        values.append(self.surveyLineData[surveyLine][searchIndex]['totalField'])
                if len(values) > 0:
                    numpyValues = numpy.array([float(record) for record in values])
                    meanValue = numpyValues.mean()
                    Gradient = self.surveyLineData[surveyLine][recordIndex]['totalField'] - meanValue
                    if abs(Gradient) < sensorNoise:
                        Gradient = 0
                    self.surveyLineData[surveyLine][recordIndex]['Gradient'] = Gradient
                else:
                    self.surveyLineData[surveyLine][recordIndex]['Gradient'] = 0

    def calculateSurveyLineGradients(self):
        
        i = 0
        for surveyLine in self.surveyLineData:
            i += len(self.surveyLineData[surveyLine]) - 1

        numpyGradientData = numpy.zeros(i, dtype={'names':['easting', 'northing', 'Gradient', 'lineNumber'], 'formats':['f8', 'f8', 'f8', 'a32']})

        index = 0
        for surveyLine in self.surveyLineData:
            multiplier = 0.0
            survey_line_len = len(self.surveyLineData[surveyLine])
            index_start = index
            index_stop = 0
            for recordIndex in range(1, survey_line_len):
                item2 = self.surveyLineData[surveyLine][recordIndex]
                item1 = self.surveyLineData[surveyLine][recordIndex - 1]
                difference = item2['totalField'] - item1['totalField']
                distance = math.sqrt(math.pow(item2['easting'] - item1['easting'], 2) + math.pow(item2['northing'] - item1['northing'], 2))

                numpyGradientData[index]['easting'] = (item2['easting'] + item1['easting']) / 2.0
                numpyGradientData[index]['northing'] = (item2['northing'] + item1['northing']) / 2.0
                numpyGradientData[index]['Gradient'] = difference / distance
                numpyGradientData[index]['lineNumber'] = item2['lineNumber']

                index += 1
                index_stop = index
                multiplier += distance

            multiplier = multiplier / (survey_line_len - 1)

            #arcpy.AddMessage('surveyLine: ' + str(surveyLine) + ', length: ' + str(survey_line_len) + ', multiplier: ' + str(multiplier))
            for recordIndex in range(index_start, survey_line_len - 1):
                numpyGradientData[recordIndex]['Gradient'] = numpyGradientData[recordIndex]['Gradient'] * multiplier
                
        return numpyGradientData

    def createNumpyArray(self, includeGradient):
        """Creates a numpy structured array from the surveyLineData dictionary
        Arguments: 
        includeGradient - bool - Set to true to include Gradient key
        Returns: 
        numpy structured array
        """
        data=[]				
        for surveyLine in self.surveyLineData:
            for record in self.surveyLineData[surveyLine]:
                data.append(record)
        if includeGradient:
            numpyData = numpy.zeros(len(data), dtype={'names':['easting', 'northing', 'totalField', 'signalQ','lineNumber', 'altitude', 'Gradient'], 'formats':['f8', 'f8', 'f8', 'i4', 'a32', 'i4', 'i4']})
            for index in range(len(data)):
                numpyData[index]['easting'] = data[index]['easting']
                numpyData[index]['northing'] = data[index]['northing']
                numpyData[index]['totalField'] = data[index]['totalField']
                numpyData[index]['signalQ'] = data[index]['signalQ']
                numpyData[index]['lineNumber'] = data[index]['lineNumber']
                numpyData[index]['altitude'] = int(round(data[index]['altitude']))
                numpyData[index]['Gradient'] = int(round(data[index]['Gradient']))

        else:
            numpyData = numpy.zeros(len(data), dtype={'names':['easting', 'northing', 'totalField','signalQ','residualAnomaly', 'spatialGradient','lineNumber', 'date', 'time', 'altitude'], 'formats':['f8', 'f8', 'f8','i4', 'f8', 'f8', 'a32', 'U10', 'U12', 'i4']})
            for index in range(len(data)):
                numpyData[index]['easting'] = data[index]['easting']
                numpyData[index]['northing'] = data[index]['northing']
                numpyData[index]['totalField'] = data[index]['totalField']
                numpyData[index]['signalQ'] = data[index]['signalQ']
                numpyData[index]['residualAnomaly'] = data[index]['residualAnomaly']
                numpyData[index]['spatialGradient'] = data[index]['spatialGradient']
                numpyData[index]['lineNumber'] = data[index]['lineNumber']
                numpyData[index]['date'] = data[index]['date']
                numpyData[index]['time'] = data[index]['time']
                numpyData[index]['altitude'] = int(round(data[index]['altitude']))		

        return numpyData


    def readDataTableWithAltitude(self, inputTable, eastingField,
                                  northingField, totalFieldField, signalQField, residualAnomalyField, spatialGradientField, lineNumberField, dateField, timeField, altitudeField,
                                  altitudeUnit):
        """Reads a table with an arcpy search cursor and populates a dictionary
        for recorded magnetometer datasets using altitude.
        Arguments: 
        inputTable - string - path to a table
        eastingField - string - easting field name
        northingField - string - northing field name
        totalFieldField - string - recorded total field name
        signalQField - int - recorded signal quality 
        lineNumberField - string - line number field name
        altitudeField - string - altitude field name
        altitudeUnit - string - altitude linear unit, either meter or feet
        """
        data = []
        unitConversions = {'meter':100, 'feet':30.48}
        with arcpy.da.SearchCursor(inputTable, [eastingField,
                                                northingField,
                                                totalFieldField,
                                                signalQField,
                                                residualAnomalyField, 
                                                spatialGradientField,
                                                lineNumberField,
                                                dateField,
                                                timeField,
                                                altitudeField]) as cursor:
            for row in cursor:
                data.append({'easting':row[0],
                             'northing':row[1],
                             'totalField':row[2],
                             'signalQ':row[3],
                             'residualAnomaly':row[4],
                             'spatialGradient':row[5],
                             'lineNumber':row[6],
                             'date':row[7],
                             'time':row[8],                    
                             'altitude':row[9] * unitConversions[altitudeUnit]})
        self.createSurveyLineData(data)

    def readDataTableWithOutAltitude(self, inputTable, eastingField,
                                     northingField, totalFieldField, signalQField, residualAnomalyField, spatialGradientField, lineNumberField, dateField, timeField, sensorAltitude):
        """Reads a table with an arcpy search cursor and populates a dictionary
        for recorded magnetometer datasets without using altitude.
        Arguments: 
        inputTable - string - path to a table
        eastingField - string - easting field name
        northingField - string - northing field name
        totalFieldField - string - recorded total field name
        signalQField - int - recorded signal quality 
        lineNumberField - string - line number field name
        sensorAltitude - double - constant sensor altitude in centimeters
        """
        data = []
        unitConversions = {'meter':100, 'feet':30.48}
        with arcpy.da.SearchCursor(inputTable, [eastingField,
                                                northingField,
                                                totalFieldField,
                                                signalQField,
                                                residualAnomalyField, 
                                                spatialGradientField,
                                                lineNumberField, 
                                                dateField, 
                                                timeField]) as cursor:
            for row in cursor:
                data.append({'easting':row[0],
                             'northing':row[1],
                             'totalField':row[2],
                             'signalQField':row[3],
                             'residualAnomaly':row[4],
                             'spatialGradient':row[5],
                             'lineNumber':row[6],
                             'date':row[7],
                             'time':row[8],
                             'altitude':sensorAltitude})
        self.createSurveyLineData(data)

    def readSurveyLinesFromFeatureClass(self, inputFeatureClass):

        data = []
        with arcpy.da.SearchCursor(inputFeatureClass, ['SHAPE@X', 'SHAPE@Y', 'totalField', 'signalQ','residualAnomaly', 'spatialGradient', 'lineNumber', 'date', 'time', 'altitude']) as cursor:
            for row in cursor:
                data.append({'easting':row[0], 
                             'northing':row[1], 
                             'totalField':row[2],
                             'signalQ':row[3],
                             'residualAnomlay':row[4],
                             'spatialGradient':row[5],
                             'lineNumber':row[6],
                             'date':row[7],
                             'time':row[8],
                             'altitude':row[9]})
        self.createSurveyLineData(data)

class Toolbox(object): 
    def __init__(self): 
        """Define the toolbox (the name of the toolbox is the name of the .pyt file)."""
        self.label = 'Magnetometer Survey Toolbox (MagTool)'
        self.alias = '' 

        # List of tool classes associated with this toolbox 
        self.tools = [InputTool, CoverageModelTool, ContouredSurfaceTool, SurveyAreaTool, DeleteIdenticalTool, ProcessRawNavigationTool, ProcessRawAltitudeTool, RawFileAssessmentTool, ResidualFieldProcessingTool, SpatialGradientProcessingTool, CreateNewProcesseingProjectTool, ConvertRawFileSchema, ConvertGradiometerFile] 

class InputTool(object):
    def __init__(self):
        self.label = 'a. Import Processed File'    
        self.description = ''
        self.category = "d. Visualization"
        self.canRunInBackground = True

    def getParameterInfo(self):

        param0 = arcpy.Parameter(displayName = 'Data Table', 
                                 name = 'inputTable', 
                                 datatype = 'Table', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param1 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEasting', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param2 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthing', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Total Field', 
                                 name = 'inputtotalField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param4 = arcpy.Parameter(displayName = 'Magnetic Signal Quality', 
                                 name = 'inputSignalQ', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param5 = arcpy.Parameter(displayName = 'Residual Anomaly Field', 
                                 name = 'inputResidual', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Spatial Gradient Field', 
                                 name = 'inputGradient', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 =  arcpy.Parameter(displayName = 'Altitude values included in table?', 
                                  name = 'useAltitudeField', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 
        
        param8 = arcpy.Parameter(displayName = 'Altitude Field', 
                                 name = 'inputAltitudeField', 
                                 datatype = 'Field', 
                                 parameterType = 'Optional', 
                                 direction = 'Input')

        param9 = arcpy.Parameter(displayName = 'Altitude is in Feet (otherwise in meters)', 
                                 name = 'useFeetAsAltitude', 
                                 datatype = 'GPBoolean', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param10 = arcpy.Parameter(displayName = 'Enter estimated altitude of sensor', 
                                  name = 'inputSensorAltitude', 
                                  datatype = 'Double', 
                                  parameterType = 'Optional', 
                                  direction = 'Input') 

        param11 = arcpy.Parameter(displayName= 'Line File Name Field', 
                                 name = 'inputLineNumber', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
                            
        param12 = arcpy.Parameter(displayName= 'Date Field', 
                                 name = 'inputDate', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param13 = arcpy.Parameter(displayName= 'Time Field', 
                                 name = 'inputTime', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param14 = arcpy.Parameter(displayName = 'Spatial Reference', 
                                 name = 'inputSpatialReference', 
                                 datatype = 'Spatial Reference', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param15 = arcpy.Parameter(displayName = 'Workspace',
                                  name = 'inputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param16 = arcpy.Parameter(displayName = 'Append rawPoints to Previous?', 
                                 name = 'yesAppend', 
                                 datatype = 'GPBoolean', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        # Default values 
        param1.value = "Field1"
        param2.value = "Field2"
        param3.value = "Field3"
        param4.value = "Field4"
        param5.value = "Field5"
        param6.value = "Field6"
        param7.value = True
        param8.value = "Field7"
        param9.value = True
        param10.value = 3
        param11.value = "Field8"
        param12.value = "Field9"
        param13.value = "Field10"
        param14.value = "NAD 1983 UTM Zone 17N"
        param16.value = False

        param1.parameterDependencies = [param0.name] 
        param2.parameterDependencies = [param0.name] 
        param3.parameterDependencies = [param0.name] 
        param4.parameterDependencies = [param0.name]
        param5.parameterDependencies = [param0.name]
        param6.parameterDependencies = [param0.name]
        param8.parameterDependencies = [param0.name] 
        param11.parameterDependencies = [param0.name]
        param12.parameterDependencies = [param0.name]
        param13.parameterDependencies = [param0.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11, param12, param13, param14, param15, param16] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters):
        if parameters[7].value == True:
            parameters[8].enabled = True
            parameters[10].enabled = False

        if parameters[7].value == False:
            parameters[8].enabled = False
            parameters[10].enabled = True

        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input parameters 
        inputTable = parameters[0].valueAsText 
        arcpy.AddMessage('Table: {}'.format(inputTable))

        inputEastingField = parameters[1].valueAsText 
        arcpy.AddMessage('Easting: {}'.format(inputEastingField))

        inputNorthingField = parameters[2].valueAsText 
        arcpy.AddMessage('Northing: {}'.format(inputNorthingField)) 

        inputGammaField = parameters[3].valueAsText 
        arcpy.AddMessage('Total Field: {}'.format(inputGammaField)) 

        inputSignalQ = parameters[4].valueAsText 
        arcpy.AddMessage('Siqnal Quality: {}'.format(inputSignalQ)) 

        inputResidualField = parameters[5].valueAsText
        arcpy.AddMessage('Residual Anomaly Field: {}'.format(inputResidualField)) 

        inputSpatialGradientField = parameters[6].valueAsText
        arcpy.AddMessage('Spatial Gradient Field: {}'.format(inputSpatialGradientField)) 

        useAltitudeField = parameters[7].value 
        arcpy.AddMessage('Use altitude field? {}'.format(str(useAltitudeField))) 
        
        if useAltitudeField:
            inputAltitudeField = parameters[8].valueAsText 
            arcpy.AddMessage('Altitude Field: {}'.format(inputAltitudeField))
        
        useFeetAsAltitude = parameters[9].value 
        arcpy.AddMessage('Altitude is measured in feet? {}'.format(str(useFeetAsAltitude))) 
        
        if useAltitudeField == False:
            unitConversions = {'meter':100, 'feet':30.48}
            if useFeetAsAltitude:
                inputSensorAltitude = parameters[10].value * unitConversions['feet']
                arcpy.AddMessage('Input sensor altitude: {}'.format(str(inputSensorAltitude)))
            else:
                inputSensorAltitude = parameters[10].value * unitConversions['meter']
                arcpy.AddMessage('Input sensor altitude: {}'.format(str(inputSensorAltitude)))

        inputLineNumberField = parameters[11].valueAsText 
        arcpy.AddMessage('Line Number Field: {}'.format(inputLineNumberField)) 

        inputDateField = parameters[12].valueAsText 
        arcpy.AddMessage('Date Field: {}'.format(inputDateField))

        inputTimeField = parameters[13].valueAsText 
        arcpy.AddMessage('Time Field: {}'.format(inputTimeField))

        inputSpatialReference = parameters[14].value 
        arcpy.AddMessage('Spatial reference: {}'.format(inputSpatialReference.name)) 

        outputWorkspace = parameters[15].valueAsText

        arcpy.AddMessage('Workspace: {}'.format(outputWorkspace)) 

        appendrawPoints = parameters[16].value 
        arcpy.AddMessage('Append rawPoints? {}'.format(str(appendrawPoints)))

        # Set output workspace and overwrite 

        arcpy.env.workspace = r'{}'.format(outputWorkspace)
        arcpy.env.overwriteOutput = True

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        # Read the data from the table 
        if useAltitudeField: 
            if useFeetAsAltitude: 
                gammaReadings.readDataTableWithAltitude(inputTable, inputEastingField,
                                                        inputNorthingField, inputGammaField, inputSignalQ, inputResidualField, inputSpatialGradientField,
                                                        inputLineNumberField, inputDateField, inputTimeField, inputAltitudeField, 'feet')
            else: 
                gammaReadings.readDataTableWithAltitude(inputTable, inputEastingField,
                                                        inputNorthingField, inputGammaField, inputSignalQ, inputResidualField, inputSpatialGradientField,
                                                        inputLineNumberField, inputDateField, inputTimeField, inputAltitudeField, 'meter') 
        else:
            if useFeetAsAltitude:
                gammaReadings.readDataTableWithOutAltitude(inputTable,
                                                           inputEastingField, inputNorthingField, inputGammaField, inputSignalQ, inputResidualField, inputSpatialGradientField,
                                                           inputLineNumberField, inputDateField, inputTimeField, inputSensorAltitude)
            else:
                gammaReadings.readDataTableWithOutAltitude(inputTable,
                                                           inputEastingField, inputNorthingField, inputGammaField, inputSignalQ, inputResidualField, inputSpatialGradientField,
                                                           inputLineNumberField, inputDateField, inputTimeField, inputSensorAltitude)
        arcpy.AddMessage('Read the data table')

        # copy rawPoints features if it exists to proceed appending
        if appendrawPoints == True:
            if arcpy.Exists("rawPoints"):
                arcpy.Rename_management("rawPoints","rawPointsapp")
                arcpy.management.Delete("rawPoints")
        if appendrawPoints == False:
            if arcpy.Exists("rawPoints"): 
                arcpy.management.Delete("rawPoints")

        arcpy.AddMessage('Cleaned Workspace')

        # Create a feature class from the gamma readings 
        gammaNumpyArray = gammaReadings.createNumpyArray(False) 

        out_table = "{}\{}".format(outputWorkspace,"rawPoints")
        arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, out_table, ['easting', 'northing'], inputSpatialReference) 
        arcpy.AddMessage('Created rawPoints') 

        #append points
        if arcpy.Exists("rawPointsapp"):
            arcpy.management.Append("rawPoints", "rawPointsapp", "TEST")
            arcpy.AddMessage('Appending...')
            arcpy.management.Delete("rawPoints")
            arcpy.management.CopyFeatures("rawPointsapp", "rawPoints")
            arcpy.management.Delete("rawPointsapp")
            arcpy.AddMessage('Appended rawPoints')

        arcpy.env.workspace = outputWorkspace
        # convert points to lines >> surveyLines
        arcpy.PointsToLine_management('rawPoints', 'surveyLines', 'lineNumber')
        arcpy.AddMessage('Created surveyLines')

        #convert rawPoints into df to get summary values of totalField
        qcTFstats=arcpy.da.TableToNumPyArray("rawPoints",('totalField'))
        df=pd.DataFrame(qcTFstats)
        meanTotalField=df.round(1)['totalField'].mean()
        arcpy.AddMessage("Mean Total Field Value: {0} nT".format(meanTotalField))
        maxTotalField=df.round(1)['totalField'].max()
        arcpy.AddMessage("Maximum Total Field Value: {0} nT".format(maxTotalField))
        minTotalField=df.round(1)['totalField'].min()
        arcpy.AddMessage("Minimum Total Field Value: {0} nT".format(minTotalField))

        surveyDist=arcpy.da.TableToNumPyArray("surveyLines",('Shape_Length'))
        df2=pd.DataFrame(surveyDist)
        totalsurveydist=((df2.round(2)['Shape_Length'].sum())/1000)
        arcpy.AddMessage("Total Survey Runline Distance: {0} km".format(totalsurveydist))

        arcpy.AddMessage("Input Tool Complete")       

class CoverageModelTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class."""
        self.label = 'a. Coverage Mapping'
        self.description = ''
        self.category = "e. Survey Coverage Assessment" 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""

        param0 = arcpy.Parameter(displayName = 'Input Survey Points Feature Class', 
                                 name = 'inputSurveyPoints', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param1 = arcpy.Parameter(displayName = 'Survey Area', 
                                  name = 'inputSurveyArea', 
                                  datatype = 'Feature Class', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param2 = arcpy.Parameter(displayName = 'Magnetic Moment', 
                                  name = 'inputMagneticMoment', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Minimum Anomaly Amplitude', 
                                  name = 'inputSensorNoise', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Mass Threshold', 
                                  name = 'inputMassThreshold', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param5 = arcpy.Parameter(displayName = 'Points back to include in neighborhood for spatial gradient calculation', 
                                  name = 'inputGradientPointsBack', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param6 = arcpy.Parameter(displayName = 'Points forward to include in neighborhood for spatial gradeint calculation', 
                                  name = 'inputGradientPointsForward', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param7 = arcpy.Parameter(displayName = 'Cell size for output rasters', 
                                  name = 'inputCellSize', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param8 = arcpy.Parameter(displayName = 'Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param2.value = 50
        param3.value = 5
        param4.value = 100
        param5.value = 5
        param6.value = 5
        param7.value = 1

        # Dependencies 
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.CheckExtension('Spatial') != 'Available': 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
    
        return

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                workspaceparts = os.path.split(inputFeatureClass)
            
                arcpy.env.workspace = workspaceparts[0]
                desc = arcpy.Describe(inputFeatureClass)

                if (desc.shapeType != 'Point'):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be points feature class')
                else :
                    fieldList = arcpy.ListFields(inputFeatureClass)
                    b_gamma = False
                    b_line = False

                    for field in fieldList:
                        if field.name == 'totalField':
                            b_gamma = True
                        elif field.name == 'lineNumber':
                            b_line = True

                    if b_gamma == False or b_line == False:
                        parameters[0].setErrorMessage('Not all required input fields present lineNumber and totalField')

        except Exception as e:
            return
        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input parameters 
        inputFeatureClass = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputFeatureClass))

        inputSurveyArea = parameters[1].valueAsText 
        arcpy.AddMessage('Survey area: {}'.format(inputSurveyArea)) 

        inputMagneticMoment = parameters[2].value
        arcpy.AddMessage('Magnetic moment: {}'.format(str(inputMagneticMoment))) 

        inputSensorNoise = parameters[3].value
        arcpy.AddMessage('Minimum Anomaly Amplitude: {}'.format(str(inputSensorNoise))) 

        inputMassThreshold = parameters[4].value
        arcpy.AddMessage('Mass threshold: {}'.format(str(inputMassThreshold)))

        inputGradientPointsBack = int(parameters[5].value)
        arcpy.AddMessage('Points back for spatial gradient calculation: {}'.format(str(inputGradientPointsBack)))

        inputGradientPointsForward = int(parameters[6].value)
        arcpy.AddMessage('Points forward for spatial calculation: {}'.format(str(inputGradientPointsForward)))

        inputCellSize = parameters[7].value
        arcpy.AddMessage('Cell size for output rasters: {}'.format(str(inputCellSize)))

        outputWorkspace = parameters[8].valueAsText

        arcpy.AddMessage('outputWorkspace: {}'.format(outputWorkspace))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.env.mask = inputSurveyArea 
        arcpy.env.extent = inputSurveyArea

        arcpy.AddMessage('Set environment variables')         

        suffix = ''
        delta_suffix = ''
        delta_suffix_2 = ''

        # Create the logTable
        now = datetime.datetime.now()
        logTable = 'CoverageMappingLog' + suffix
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.AddField_management(logTable, 'Process', 'TEXT')
        arcpy.AddField_management(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
                
        logCursor.insertRow(('Survey area: {}'.format(inputSurveyArea), datetime.datetime.now()))
        logCursor.insertRow(('Magnetic moment: {}'.format(str(inputMagneticMoment)), datetime.datetime.now()))
        logCursor.insertRow(('Sensor noise: {}'.format(str(inputSensorNoise)), datetime.datetime.now()))
        logCursor.insertRow(('Mass threshold: {}'.format(str(inputMassThreshold)), datetime.datetime.now()))
        logCursor.insertRow(('Points back for spatial gradient calculation: {}'.format(str(inputGradientPointsBack)), datetime.datetime.now()))
        logCursor.insertRow(('Points forward for spatial gradient calculation: {}'.format(str(inputGradientPointsForward)), datetime.datetime.now()))
        logCursor.insertRow(('Cell size for output rasters: {}'.format(str(inputCellSize)), datetime.datetime.now()))

        # Check out the spatial analyst extension 
        arcpy.CheckOutExtension('Spatial') 
        arcpy.AddMessage('Checked out Spatial Analyst extension') 
        logCursor.insertRow(('Checked out Spatial Analyst extension', datetime.datetime.now()))

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        # Read the data from the table 
        gammaReadings.readSurveyLinesFromFeatureClass(inputFeatureClass)

        arcpy.AddMessage('Read the data table')
        logCursor.insertRow(('Read the data table', datetime.datetime.now()))

        # Calculate spatial gradient values 
        gammaReadings.calculateGradient(inputSensorNoise, inputGradientPointsBack, inputGradientPointsForward)
        arcpy.AddMessage('Calculated Spatial Gradient') 
        logCursor.insertRow(('Calculated Spatial Gradient', datetime.datetime.now()))

        inputSpatialReference = arcpy.Describe(inputFeatureClass).spatialReference

        # Create a feature class from the gamma readings 
        gammaNumpyArray = gammaReadings.createNumpyArray(True)
        intGradient =  'integerGradient' + delta_suffix
        outputFeatureClass = os.path.join(outputWorkspace, intGradient)
        if arcpy.Exists(outputFeatureClass):
            arcpy.management.Delete(outputFeatureClass)
        arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, outputFeatureClass, ['easting', 'northing'], inputSpatialReference) 
        arcpy.AddMessage('Created Gradient Integer Output') 
        logCursor.insertRow(('Created Gradient Integer Output', datetime.datetime.now()))

        del gammaNumpyArray
        del gammaReadings

        # Allocate Gradient Values as a raster 
        arcpy.AddMessage('Set Output Extent')
        integerGradientSurface = EucAllocation(intGradient, '', '', inputCellSize, 'Gradient') 
##        spatialGradientSurface.save('integerGradient' + delta_suffix_2) 
        arcpy.AddMessage('Created Confidence Model Integer Gradient Surface')
        logCursor.insertRow(('Created Confidence Model Integer Gradient Surface', datetime.datetime.now()))

        # Create the absolute value of spatialGradient
        absintegerGradientSurface = Abs(integerGradientSurface) 
        absintegerGradientSurface.save('absintegerGradient' + delta_suffix_2) 
        arcpy.AddMessage('Created abs Confidence Model Integer Gradient Surface') 
        logCursor.insertRow(('Created abs Confidence Model Integer Gradient Surface', datetime.datetime.now()))

        # Set the geoprocessing snap raster to absintegerGradient 
        arcpy.env.snapRaster = 'absintegerGradient' + delta_suffix_2
        arcpy.AddMessage('Set snap raster') 
        logCursor.insertRow(('Set snap raster', datetime.datetime.now()))

        # Allocate altitude values as a raster 
        altitudeCentimeters = EucAllocation(intGradient, '', '', inputCellSize, 'altitude') 
##        altitudeCentimeters.save('altitudeCentimeters' + delta_suffix_2) 
        arcpy.AddMessage('Created altitude') 
        logCursor.insertRow(('Created altitude', datetime.datetime.now()))

        # Create a horizontal euclidean distance raster from the input source points 
        horizontalDistance = EucDistance(intGradient, '', inputCellSize) 

        horizontalCentimeters = Int(horizontalDistance * inputSpatialReference.metersPerUnit * 100) 
##        horizontalCentimeters.save('horizontalCentimeters' + delta_suffix_2) 
        arcpy.AddMessage('Created horizontalCentimeters') 
        logCursor.insertRow(('Created horizontalCentimeters', datetime.datetime.now()))

        # Calculate the final distance raster in centimeters 
        centimeters = Int((horizontalCentimeters**2 + altitudeCentimeters**2)**0.5) 
        centimeters.save('sensorDistance' + delta_suffix_2) 
        arcpy.AddMessage('Created centimeters with altitude') 
        logCursor.insertRow(('Created centimeters with altitude field values', datetime.datetime.now()))

        # Create mass (kg) of potentially missed objects 
        potentialMassMissed = Int((inputSensorNoise * 0.000001 * (centimeters)**3) / (inputMagneticMoment)) 
        potentialMassMissed.save('potentialMassMissed' + suffix) 
        arcpy.AddMessage('Created potentialMassMissed') 
        logCursor.insertRow(('Created potentialMassMissed', datetime.datetime.now()))

        try:
            arcpy.BuildRasterAttributeTable_management('potentialMassMissed' + suffix, "Overwrite")
        except:
            arcpy.AddMessage("Build Raster Attribute Table failed for potential mass missed.")

        # Create mass (kg) of potentially found objects 
        potentialMassObserved = Int((absintegerGradientSurface * 0.000001 * (centimeters)**3) / (inputMagneticMoment)) 
        potentialMassObserved.save('potentialMassObserved' + suffix) 
        arcpy.AddMessage('Created potentialMassObserved') 
        logCursor.insertRow(('Created potentialMassObserved', datetime.datetime.now()))

        try:
            arcpy.BuildRasterAttributeTable_management('potentialMassObserved' + suffix, "Overwrite")
        except:
            arcpy.AddMessage("Build Raster Attribute Table failed for potential mass observed.")


        # Reclassify potential mass missed raster
        if inputMassThreshold > potentialMassMissed.minimum:
            remapRange = RemapRange([[0,inputMassThreshold, 1], [inputMassThreshold, 9999999, 0]])
            confidenceResult = Reclassify(potentialMassMissed, 'Value', remapRange) 
            confidenceResult.save('confidenceResult' + suffix)
        else:
            remapRange = RemapRange([[inputMassThreshold, 9999999, 0]])
            confidenceResult = Reclassify(potentialMassMissed, 'Value', remapRange) 
            confidenceResult.save('confidenceResult' + suffix)           
        arcpy.AddMessage('Created Confidence Result')
        logCursor.insertRow(('Created Confidence Result', datetime.datetime.now()))

        # Display the percentage of the survey area adequatly surveyed
        fields = ['VALUE', 'COUNT']
        data = {r[0]:r[1] for r in arcpy.da.SearchCursor('confidenceResult' + suffix, fields)}
        keys = data.keys()
        if 0 in keys:
            if 1 in keys:
                totalCount = data[0] + data[1]
                percentCovered = (data[1] / totalCount)
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequately Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

            else:
                percentCovered = 0
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequately Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

        elif 0 not in keys:
            if 1 in keys:
                percentCovered = 1
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequately Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequately Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

        # Delete the logCursor object
        del logCursor
        arcpy.CheckInExtension('Spatial') 

class ContouredSurfaceTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)"""
        self.label = 'd. Continuous Surfaces with Contours'
        self.description = '' 
        self.category = "d. Visualization"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Points Feature Class', 
                                 name = 'inputSurveyPoints', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Total Field Channel',
                                 name = 'inputTotalField',
                                 datatype = 'Field',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Total Field Contour Interval', 
                                 name = 'inputRawContourInterval', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')       

        param3 = arcpy.Parameter(displayName = 'Create Spatial Gradient Surface?',
                                 name = 'spatialGradient',
                                 datatype = 'GPBoolean',
                                 parameterType = 'Required',
                                 direction = 'Input')
        
        param4 = arcpy.Parameter(displayName = 'Spatial Gradient Channel',
                                 name = 'inputSpatialGradient',
                                 datatype = 'Field',
                                 parameterType = 'Required',
                                 direction = 'Input')
        
        param5 = arcpy.Parameter(displayName = 'Spatial Graident Contour Interval', 
                                 name = 'inputGradientContourInterval', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')       

        param6 = arcpy.Parameter(displayName = 'Create Residual Anomaly Surface?',
                                 name = 'residualAnomaly',
                                 datatype = 'GPBoolean',
                                 parameterType = 'Required',
                                 direction = 'Input')
        
        param7 = arcpy.Parameter(displayName = 'Residual Anomaly Channel',
                                 name = 'inputResidual',
                                 datatype = 'Field',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param8 = arcpy.Parameter(displayName = 'Residual Anomaly Contour Interval', 
                                 name = 'inputResidualContourInterval', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')  

        param9 = arcpy.Parameter(displayName = 'Survey Area', 
                                 name = 'inputSurveyArea', 
                                 datatype = 'Feature Class', 
                                 parameterType = 'Required', 
                                 direction = 'Input')      
                            
        param10 = arcpy.Parameter(displayName = 'Select a Surface creation Method',
                                 name = 'surfaceMethod',
                                 datatype = 'GPString',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param11 = arcpy.Parameter(displayName = 'Min Curvature Weight', 
                                 name = 'splineWeight', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')
        
        param12 = arcpy.Parameter(displayName = 'Min Curvature Points', 
                                 name = 'splinePoints', 
                                 datatype = 'Long', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param13 = arcpy.Parameter(displayName = 'Cell Size for output rasters',
                                 name = 'inputCellSize',
                                 datatype = 'GPLong',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param14 = arcpy.Parameter(displayName = 'Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param1.value = "Field1"
        param2.value = 25.0
        param3.value = True
        param4.value = "Field2"
        param5.value = 5.0
        param6.value = True
        param7.value = "Field3"
        param8.value = 5.0
        param11.value = 0.1
        param12.value = 12.0

        param1.parameterDependencies = [param0.name] 
        param4.parameterDependencies = [param0.name] 
        param7.parameterDependencies = [param0.name]

        param10.filter.type = "ValueList"

        if arcpy.CheckExtension('3D') != "Available":
            param10.filter.list = ["Minimum Curvature"]
            param10.value = "Minimum Curvature"
        else:
            param10.filter.list = ["Minimum Curvature", "TIN"]
            param10.value = "TIN"

        # Dependencies 
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11, param12, param13, param14] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.CheckExtension('Spatial') != "Available": 
                raise Exception
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""
        
        if parameters[3].value == True:
            parameters[4].enabled = True
            parameters[5].enabled = True
        else: 
            parameters[4].enabled = False
            parameters[5].enabled = False 

        if parameters[6].value == True:
            parameters[7].enabled = True
            parameters[8].enabled = True
        else: 
            parameters[7].enabled = False
            parameters[8].enabled = False

        if (parameters[2].value < 0) :
            parameters[2].value = abs(parameters[2].value)

        if (parameters[5].value < 0) :
            parameters[5].value = abs(parameters[5].value)

        if (parameters[8].value < 0) :
            parameters[8].value = abs(parameters[8].value)

        if (parameters[10].value == "Minimum Curvature") :
            parameters[11].enabled = True
            parameters[12].enabled = True
        else :
            parameters[11].enabled = False
            parameters[12].enabled = False

        return

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                desc = arcpy.Describe(inputFeatureClass)

                if (desc.shapeType != 'Point'):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be points feature class')
                else :
                    fieldList = arcpy.ListFields(inputFeatureClass)
                    b_gamma = False
                    b_line = False

                    for field in fieldList:
                        if field.name == 'totalField':
                            b_gamma = True
                        elif field.name == 'lineNumber':
                            b_line = True

                    if b_gamma == False or b_line == False:
                        parameters[0].setErrorMessage('Not all required input fields present lineNumber and totalField')

        except Exception as e:
            return

        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input variables 
        inputFeatureClass = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputFeatureClass))

        inputTotalField = parameters[1].valueAsText
        arcpy.AddMessage('Total Field Channel; {0}'.format(inputTotalField))

        inputRawContourInterval = parameters[2].value
        arcpy.AddMessage('Total Field Contour Interval: ' + str(inputRawContourInterval))

        inputSpatialGradient = parameters[4].valueAsText
        inputGradientContourInterval = parameters[5].value
        
        createSpatialGradient = parameters[3].value
        if createSpatialGradient ==True:
            arcpy.AddMessage("Spatial Gradient Surface will be Generated") 
            arcpy.AddMessage('Spatial Gradient Field: {0}'.format(inputSpatialGradient))            
            arcpy.AddMessage('Spatial Gradient Countour Interval: {0}'.format(str(inputGradientContourInterval)))
        else: 
            arcpy.AddMessage("Spatial Gradient Surface will not be Generated")

        
        inputResidual = parameters[7].valueAsText
        inputResidualContourInterval = parameters[8].value 
        createResidualAnomaly = parameters[6].value
        if createResidualAnomaly ==True:
            arcpy.AddMessage('Residual Anomaly Surface will be Generated')            
            arcpy.AddMessage('Residual Anomaly Field: {0}'.format(inputResidual))            
            arcpy.AddMessage('Residual Gradient Contour Interval: {0}'.format(str(inputResidualContourInterval)))
        else: 
            arcpy.AddMessage('Residual Anomaly Surface will not be Generated')     

        inputSurveyArea = parameters[9].valueAsText 
        arcpy.AddMessage('inputSurveyArea: ' + inputSurveyArea)

        inputSurfaceMethod = parameters[10].value
        arcpy.AddMessage('Surface Creation Method: ' + inputSurfaceMethod)

        if (inputSurfaceMethod == "Minimum Curvature") :
            splineWeight = parameters[11].value
            splinePoints = parameters[12].value
            arcpy.AddMessage('Weight: ' + str(splineWeight))
            arcpy.AddMessage('Points: ' + str(splinePoints))

        inputCellSize = parameters[13].value
        arcpy.AddMessage('inputCellSize: ' + str(inputCellSize))

        outputWorkspace = parameters[14].valueAsText
        arcpy.AddMessage("Output Workspace: {0}".format(outputWorkspace))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.env.extent = inputSurveyArea
        arcpy.AddMessage('Scratch Workspace Folder: ' + arcpy.env.scratchFolder)
        arcpy.AddMessage('Set environment variables') 

        suffix = ''
        suffix_w_raw_contour = ''
        suffix_w_gradient_contour = ''
        suffix_w_residual_contour = ''
        suffix_w_all_contour = ''

        if (inputSurfaceMethod == "Minimum Curvature") :
            suffix = '_MinCurv_PT_' + str(int(splinePoints)) + '_CS_' + str(int(inputCellSize))
        elif (inputSurfaceMethod == "TIN") :
            suffix = '_TIN_CS_' + str(int(inputCellSize))

        suffix_w_raw_contour = suffix + '_TF_' + str(int(inputRawContourInterval))
        suffix_w_gradient_contour = suffix + '_SG_' + str(int(inputGradientContourInterval))
        suffix_w_residual_contour = suffix + '_RA_' + str(int(inputResidualContourInterval)) 
        suffix_w_all_contour = suffix + '_TF_' +str(int(inputRawContourInterval)) + '_SC_' + str(int(inputGradientContourInterval)) + '_RA_' + str(int(inputResidualContourInterval)) 

        # Create the logTable
        logTable = 'VisualizationLog' + suffix_w_all_contour
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.AddField_management(logTable, 'Process', 'TEXT')
        arcpy.AddField_management(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        logCursor.insertRow(('Survey Points: {}'.format(inputFeatureClass), datetime.datetime.now()))
        logCursor.insertRow(('Survey area: {}'.format(inputSurveyArea), datetime.datetime.now()))
        logCursor.insertRow(('Surface Generation Method: ' + inputSurfaceMethod, datetime.datetime.now()))

        if (inputSurfaceMethod == "Minimum Curvature") :
            logCursor.insertRow(('Spline Weight: {}'.format(str(splineWeight)), datetime.datetime.now()))
            logCursor.insertRow(('Spline Points: {}'.format(str(splinePoints)), datetime.datetime.now()))

        logCursor.insertRow(('Total Field Contour interval: {}'.format(str(inputRawContourInterval)), datetime.datetime.now()))
        logCursor.insertRow(('Spatial Gradient Contour interval: {}'.format(str(inputGradientContourInterval)), datetime.datetime.now()))
        logCursor.insertRow(('Residual Anomaly Contour interval: {}'.format(str(inputResidualContourInterval)), datetime.datetime.now()))
        logCursor.insertRow(('Cell size for output rasters: {}'.format(str(inputCellSize)), datetime.datetime.now()))

        # Check out the spatial analyst extension 
        arcpy.CheckOutExtension('Spatial') 
        arcpy.AddMessage('Checked out Spatial Analyst extension')
        logCursor.insertRow(('Checked out Spatail Analyst extension', datetime.datetime.now())) 

        # Interpolate a surface of total field values
        if (inputSurfaceMethod == "Minimum Curvature") :
            # Set the geoprocessing mask 
            arcpy.env.mask = inputSurveyArea 
            arcpy.AddMessage('Set analysis mask')
            logCursor.insertRow(('Set analysis mask', datetime.datetime.now()))
            
            totalFieldSurface = 'totalField' + suffix
            totalField = Spline(inputFeatureClass, inputTotalField, str(inputCellSize), 'REGULARIZED', splineWeight, splinePoints)
            totalField.save(totalFieldSurface)		
            arcpy.AddMessage('Interpolated total field surface by Minimun Curvature') 
            logCursor.insertRow(('Interpolated total field by Minimum Curvature', datetime.datetime.now()))

            if createSpatialGradient == True: 
                spatialGradientSurface = 'spatialGradient' + suffix
                spatialGradient = Spline(inputFeatureClass, inputSpatialGradient, str(inputCellSize), 'REGULARIZED', splineWeight, splinePoints)
                spatialGradient.save(spatialGradientSurface)		
                arcpy.AddMessage('Interpolated spatial gradient surface by Minimum Curvature') 
                logCursor.insertRow(('Interpolated spatial gradient by Minimum Curvature', datetime.datetime.now()))

            else: 
                arcpy.AddMessage("--------------------------------------------")
                arcpy.AddMessage('Spatial Gradient surface not generated ')
                arcpy.AddMessage("--------------------------------------------")

            if createResidualAnomaly == True: 
                residualAnomalySurface = 'residualAnomaly' + suffix
                residualAnomaly = Spline(inputFeatureClass, inputResidual, str(inputCellSize), 'REGULARIZED', splineWeight, splinePoints )
                residualAnomaly.save(residualAnomalySurface)
                arcpy.AddMessage('Interpolated residual anomlay surface by Minimum Curvature')
                logCursor.insertRow(('Interpolated residual anomlay by Minimum Curvature', datetime.datetime.now()))

            else: 
                arcpy.AddMessage("--------------------------------------------")
                arcpy.AddMessage('Residual Anomlay surface not generated')
                arcpy.AddMessage("--------------------------------------------")

        elif (inputSurfaceMethod == "TIN") :
            arcpy.CheckOutExtension("3D")
            arcpy.AddMessage('Checked out 3D extension') 
            logCursor.insertRow(('Checked out 3D extension', datetime.datetime.now()))

            try:
                desc = arcpy.Describe(inputFeatureClass)
                inputSpatialReference = desc.spatialReference
                rawTIN = arcpy.env.scratchFolder + '\\rawTin'
                arcpy.AddMessage('Creating TIN dataset: ' + rawTIN)
                arcpy.ddd.CreateTin(rawTIN, inputSpatialReference, inputFeatureClass + " {0} masspoints".format(inputTotalField), "DELAUNAY") 
                arcpy.AddMessage('TIN dataset created')
                logCursor.insertRow(('TIN dataset created', datetime.datetime.now()))

                totalFieldSurface = 'totalField' + suffix
                totalFieldSurfaceTemp = totalFieldSurface + '_temp' 
                arcpy.env.mask = inputSurveyArea 
                arcpy.TinRaster_3d(rawTIN, totalFieldSurfaceTemp, "FLOAT", "LINEAR", "CELLSIZE " + str(inputCellSize), "1")
                
                rGS = ExtractByMask(totalFieldSurfaceTemp, inputSurveyArea)
                rGS.save(totalFieldSurface)
                
                arcpy.Delete_management(totalFieldSurfaceTemp)
                arcpy.AddMessage('TIN surface raster created')

                arcpy.Delete_management(rawTIN)
                arcpy.AddMessage('TIN deleted:' + rawTIN)

                
                if createSpatialGradient == True:
                    gradientTIN = arcpy.env.scratchFolder + '\\gradientTin'
                    arcpy.AddMessage('Creating TIN dataset: ' + gradientTIN)
                    arcpy.ddd.CreateTin(gradientTIN, inputSpatialReference, inputFeatureClass + " {0} Mass_Points".format(inputSpatialGradient), "DELAUNAY") 
                    arcpy.AddMessage('TIN dataset created')
                    logCursor.insertRow(('TIN dataset created', datetime.datetime.now()))

                    spatialGradientSurface = 'spatialGradient' + suffix
                    spatialGradientSurfaceTemp = spatialGradientSurface + '_temp' 
                    arcpy.env.mask = inputSurveyArea 
                    arcpy.TinRaster_3d(gradientTIN,  spatialGradientSurfaceTemp, "FLOAT", "LINEAR", "CELLSIZE " + str(inputCellSize), "1")
                    
                    temp = ExtractByMask(spatialGradientSurfaceTemp, inputSurveyArea)
                    temp.save(spatialGradientSurface)
                    
                    arcpy.Delete_management(spatialGradientSurfaceTemp)
                    arcpy.AddMessage('TIN surface raster created')

                    arcpy.Delete_management(gradientTIN)
                    arcpy.AddMessage('TIN deleted:' + gradientTIN)
                else: 
                    arcpy.AddMessage("--------------------------------------------")
                    arcpy.AddMessage('Spatial Gradient surface not generated')
                    arcpy.AddMessage("--------------------------------------------")

                
                if createResidualAnomaly == True:
                    residualTIN = arcpy.env.scratchFolder + '\\residualTin'
                    arcpy.AddMessage('Creating TIN dataset: ' + residualTIN)
                    arcpy.ddd.CreateTin(residualTIN, inputSpatialReference, inputFeatureClass + " {0} Mass_Points".format(inputResidual), "DELAUNAY") 
                    arcpy.AddMessage('TIN dataset created')
                    logCursor.insertRow(('TIN dataset created', datetime.datetime.now()))
                    
                    residualAnomalySurface = 'residualAnomaly' + suffix
                    residualAnomalySurfaceTemp = residualAnomalySurface + '_temp' 
                    arcpy.env.mask = inputSurveyArea 
                    arcpy.TinRaster_3d(residualTIN,  residualAnomalySurfaceTemp, "FLOAT", "LINEAR", "CELLSIZE " + str(inputCellSize), "1")
                    
                    temp = ExtractByMask(residualAnomalySurfaceTemp, inputSurveyArea)
                    temp.save(residualAnomalySurface)
                    
                    arcpy.Delete_management(residualAnomalySurfaceTemp)
                    arcpy.AddMessage('TIN surface raster created')

                    arcpy.Delete_management(residualTIN)
                    arcpy.AddMessage('TIN deleted:' + residualTIN)
                else: 
                    arcpy.AddMessage("--------------------------------------------")
                    arcpy.AddMessage('Residual Anomaly surface not generated')
                    arcpy.AddMessage("--------------------------------------------")

            except arcpy.ExecuteError as e:
                arcpy.AddError('Tool Error Creating TIN dataset: ' + e.message)
                logCursor.insertRow(('tool error creating TIN dataset: ' + e.message, datetime.datetime.now()))
                return
            except Exception as e:
                arcpy.AddError('Error Creating TIN dataset: ' + e.message)
                logCursor.insertRow(('error creating TIN dataset: ' + e.message, datetime.datetime.now()))
                return
            arcpy.CheckInExtension("3D")
        else :
            arcpy.AddMessage('Unknown interpolation method: ' + inputSurfaceMethod)
            logCursor.insertRow(('Unknown Interpolation Method', datetime.datetime.now()))
            return

        arcpy.env.workspace = outputWorkspace 

        # Create contours on each continuous surface based on user input intervals 
        if arcpy.Exists(totalFieldSurface):
            Contour(totalFieldSurface, 'totalField' + suffix_w_raw_contour, inputRawContourInterval) 
            arcpy.AddMessage('Created total field contour lines')
            arcpy.AddMessage("--------------------------------------------")
            logCursor.insertRow(('Created total field contour lines', datetime.datetime.now()))
        else :
            arcpy.AddMessage('Could not find surface : ' + totalFieldSurface + ' in workspace: ' + arcpy.env.workspace)

        # Create contours of the gradient gamma values 
        if createSpatialGradient == True:
            Contour(spatialGradientSurface, 'spatialGradient' + suffix_w_gradient_contour, inputGradientContourInterval, inputGradientContourInterval / 2.0) 
            arcpy.AddMessage('Created spatial gradient contour lines')
            arcpy.AddMessage("--------------------------------------------")
            logCursor.insertRow(('Created spatial gradient contour lines', datetime.datetime.now()))
        else :
            arcpy.AddMessage('--------------------------------')

        if createResidualAnomaly == True:
            Contour(residualAnomalySurface, 'residualAnomaly' + suffix_w_residual_contour, inputResidualContourInterval)
            arcpy.AddMessage('Created residual anomaly contour lines')
            arcpy.AddMessage("--------------------------------------------")
            logCursor.insertRow(('Created residual anomaly contour lines', datetime.datetime.now()))
        else: 
            arcpy.AddMessage('--------------------------------')

        # Delete the logCursor object
        del logCursor
        arcpy.CheckInExtension('Spatial') 
        arcpy.AddMessage("--------------------------------------------")
        arcpy.AddMessage("Completed Surface Generation and Contouring")
        arcpy.AddMessage("--------------------------------------------")

class DeleteIdenticalTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'b. Delete Identical Points'
        self.description = '' 
        self.category = "d. Visualization"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Point Feature Class', 
                                 name = 'inputSurveyFeatures', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Spatial Tolerance ([dist] [unit])', 
                                 name = 'tolerance', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        # Default values 
        param1.value = '0.01 METERS'
        return [param0, param1] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        inputSurveyFeatures = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputSurveyFeatures))

        inputTolerance = parameters[1].value
        arcpy.AddMessage('Tolerance: {}'.format(inputTolerance))

        arcpy.management.DeleteIdentical(inputSurveyFeatures,"SHAPE", inputTolerance, 0)
        arcpy.AddMessage(arcpy.GetMessages())
        arcpy.AddMessage("Tool Complete")

class SurveyAreaTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'c. Generate Survey Boundary'
        self.description = '' 
        self.category = "d. Visualization"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Line or Point Feature Class', 
                                 name = 'inputSurveyFeatures', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Buffer Distance (same units as source)', 
                                 name = 'inputBufferDistance', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param2 =  arcpy.Parameter(displayName = 'Send Outputs to same geodatabase as input?', 
                                  name = 'useInputGeodatabase', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Existing Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param1.value = 5.0
        param2.value = True
        return [param0, param1, param2, param3] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""

        if (parameters[2].value == True):
            parameters[3].enabled = False

            if (parameters[0].altered and len(parameters[0].valueAsText) > 0) : 
                workspaceparts = os.path.split(parameters[0].valueAsText)
                parameters[3].value = workspaceparts[0]
            else :
                parameters[3].value = ''
        else :
            parameters[3].enabled = True

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        
        if parameters[1].value < 3.0 :
            parameters[1].setErrorMessage('Buffer distance must be 3 meters or greater')

        try:
            arcpy.env.workspace = parameters[3].valueAsText
            if len(arcpy.env.workspace) > 0 and arcpy.Exists('surveyArea'):
                parameters[0].setWarningMessage('Buffer already defined. Datasets may be overwritten')
        except Exception as e:
            return

        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                desc = arcpy.Describe(inputFeatureClass)

                if ((desc.shapeType != 'Point') and (desc.shapeType != 'Polyline')):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be point or line feature class')

        except Exception as e:
            return

        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input variables 
        inputSurveyFeatures = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputSurveyFeatures))

        inputBufferDistance = parameters[1].value
        arcpy.AddMessage('inputBufferDistance: {}'.format(str(inputBufferDistance)))

        outputWorkspace = parameters[3].valueAsText

        arcpy.AddMessage('output Geodatabase: {}'.format(outputWorkspace))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.AddMessage('Set environment variables') 

        # Create the logTable
        now = datetime.datetime.now()
        logTable = 'SurveyAreaLog_{}'.format(now.strftime('%Y%m%d%H%M%S'))
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.management.AddField(logTable, 'Process', 'TEXT')
        arcpy.management.AddField(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        logCursor.insertRow(('Survey Features: {}'.format(inputSurveyFeatures), datetime.datetime.now()))
        logCursor.insertRow(('Buffer Distance: {}'.format(inputBufferDistance), datetime.datetime.now()))

        # Make convex hull 
        arcpy.management.MinimumBoundingGeometry(inputSurveyFeatures, 'convexHull', 'CONVEX_HULL') 
        arcpy.AddMessage('Made convex hull') 
        logCursor.insertRow(('Made convex hull', datetime.datetime.now()))

        # Create survey area by buffering convex hull 
        arcpy.analysis.Buffer('convexHull', 'convexHullArea', inputBufferDistance, 'FULL', '', 'ALL') 
        arcpy.AddMessage('Made survey area by buffering convex hull')
        if arcpy.Exists('convexHull'):
            arcpy.management.Delete('convexHull')
        logCursor.insertRow(('Made survey area by buffering convex hull', datetime.datetime.now()))

        arcpy.AddMessage('Creating Blanking Distance Coverage Polygon') 

        arcpy.analysis.Buffer(inputSurveyFeatures, 'surveyArea_temp1', inputBufferDistance, 'FULL', 'ROUND', 'ALL')

        arcpy.AddMessage('bufferred individual input features') 
        logCursor.insertRow(('bufferred individual input features', datetime.datetime.now()))

##        try:
##        
##            arcpy.management.EliminatePolygonPart('surveyArea_temp1', 'surveyArea_temp2', 'PERCENT', '', 99.0, 'CONTAINED_ONLY')
##
##            arcpy.AddMessage('Eliminated internal voids in search area') 
##            logCursor.insertRow(('Eliminated internal voids in search area', datetime.datetime.now()))
##        except Exception:
##            arcpy.AddMessage('Unable to eliminate internal voids in search area') 
##            logCursor.insertRow(('Unable to eliminate internal voids in search area', datetime.datetime.now()))

        tempfeature = None

        try:
            if arcpy.Exists('surveyArea_temp2'):
                tempfeature = 'surveyArea_temp2'
            else:
                tempfeature = 'surveyArea_temp1'

            arcpy.cartography.SimplifyPolygon(tempfeature, 'coverageArea', "POINT_REMOVE", (inputBufferDistance / 4.0), 0, 'RESOLVE_ERRORS', 'NO_KEEP')

            arcpy.AddMessage('Simplified blanking distance coverage area') 
            logCursor.insertRow(('Simplified bufferred survey area', datetime.datetime.now()))
        except Exception:
            arcpy.AddMessage('Unable to simplify survey area') 
            logCursor.insertRow(('Unable to simplify survey area', datetime.datetime.now()))

            arcpy.management.Rename(tempfeature, 'coverageArea')

        if arcpy.Exists('coverageArea' + '_Pnt'):
            arcpy.management.Delete('coverageArea' + '_Pnt')

        if arcpy.Exists('surveyArea_temp1'):
            arcpy.management.Delete('surveyArea_temp1')
        
        if arcpy.Exists('surveyArea_temp2'):
            arcpy.management.Delete('surveyArea_temp2')

        logCursor.insertRow(('Made Blanking Distance Polygon', datetime.datetime.now()))

        # Delete the logCursor object
        del logCursor

class ProcessRawNavigationTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'b. Process Raw Navigation'
        self.description = '' 
        self.category = "b. QA/QC Tools"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Processing Project', 
                                 name = 'procProj', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEastingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Rolling Stats Window Size', 
                                 name = 'window', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Line Field', 
                                 name = 'inputSurveyLine', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Time Field', 
                                 name = 'inputTime', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Output File Name (tool will append _navsmth)', 
                                 name = 'outputFileName', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        # Default values 
        param2.value = "Field1"
        param3.value = "Field2"
        param4.value = 20
        param5.value = "Field4"
        param6.value = "Field5"

        param2.parameterDependencies = [param1.name]
        param3.parameterDependencies = [param1.name]
        param5.parameterDependencies = [param1.name]
        param6.parameterDependencies = [param1.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        procProj=parameters[0].valueAsText
        arcpy.AddMessage('------------------')
        arcpy.AddMessage('Processing Project: {0}'.format(procProj))
        arcpy.AddMessage('------------------')

        inputRawFile = parameters[1].valueAsText 
        arcpy.AddMessage('Raw File: {}'.format(inputRawFile))
        arcpy.AddMessage('------------------')

        Easting=parameters[2].valueAsText

        Northing=parameters[3].valueAsText

        window=(int(parameters[4].value))
        arcpy.AddMessage('Rolling Stats Window Size: {} points'.format(window))
        arcpy.AddMessage('------------------')

        inputLinefield=parameters[5].valueAsText

        inputTime=parameters[6].valueAsText

        navPlotPathway=(r'{0}\03_Processing_Figures\Navigation_Figures'.format(procProj))

        navSmthPathway=(r'{0}\02_Processing_Workspaces\NavSmth_temp'.format(procProj))

        outputWorkspace=(r'{0}\01_Data_Files\03_Smoothed_Files'.format(procProj))

        outputFileName=parameters[7].valueAsText 

        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Read raw file")
        arcpy.AddMessage('------------------')

        ##iterate a list of unique line names
        lineNameList=df['{}'.format(inputLinefield)].unique()
        arcpy.AddMessage(lineNameList)
        arcpy.AddMessage('------------------')
        arcpy.AddMessage("Total Number of Lines: {}".format(len(lineNameList)))
        arcpy.AddMessage('------------------')
        plt.rcParams.update({'figure.max_open_warning': 0})

        ## define a function to perform a rolling stats computation on single line of data
        def rollingStatsperLine(x): 
            arcpy.AddMessage("Processing Line: {}".format(lineNameList[x]))
            mask=df['{}'.format(inputLinefield)]==lineNameList[x]
            dfx=df[mask].copy()
            dfx['{}_smth'.format(Easting)]=dfx['{}'.format(Easting)].rolling(window,center=True).mean()
            dfx['{}_smth'.format(Northing)]=dfx['{}'.format(Northing)].rolling(window,center=True).mean()
            dfx.to_csv(r"{}\{}{}.txt".format(navSmthPathway,outputFileName,x),float_format='%.2f',index=False)
            
            fig = plt.figure()
            plt.title("Raw vs Smoothed Navigation",fontdict=None, loc='center', pad=None)
            plt.tick_params(axis='both', which='both', bottom=False, top=False, labelbottom=False) 
            plt.yticks([])

            ax1 = fig.add_subplot(411)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}'.format(Easting)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            plt.ticklabel_format(style='plain', useOffset=False, axis='y')
            plt.legend(['Raw X'])
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(412)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}_smth'.format(Easting)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            plt.ticklabel_format(style='plain', useOffset=False, axis='y')
            plt.legend(['Smth X'])
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(413)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}'.format(Northing)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            plt.ticklabel_format(style='plain', useOffset=False, axis='y')
            plt.legend(['Raw Y'])
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(414)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}_smth'.format(Northing)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            plt.ticklabel_format(style='plain', useOffset=False, axis='y')
            plt.legend(['Smth Y'])
            ax1.axes.get_xaxis().set_visible(False)

            plt.savefig(r"{0}\{1}.jpg".format(navPlotPathway,lineNameList[x]), dpi=300)
            plt.close(fig)
            del fig
            return

        ## call the function and run as iterations of the line name list
        x=0
        while x <= len(lineNameList):
            rollingStatsperLine(x)
            x +=1
            if x == len(lineNameList):
                break

        ## read directory of CSV files generated for each df split by line, merge into a single df
        path =(r"{}".format(navSmthPathway))
        all_files = glob.glob(os.path.join(path,"{}*.txt".format(outputFileName)))
        df_from_each_file = (pd.read_csv(f,sep=',')for f in all_files)
        df_merged = pd.concat(df_from_each_file,ignore_index=True)
        df_final=df_merged.dropna()
        arcpy.AddMessage('------------------')
        arcpy.AddMessage("Merged Results")
        arcpy.AddMessage('------------------')

        ## export merged df as full csv file
        os.chdir(r"{}".format(outputWorkspace))
        df_final.to_csv("{}_navsmth.txt".format(outputFileName),index=False)

        for f in all_files:
            os.remove(f)

        arcpy.AddMessage("completed line smoothing")

class ProcessRawAltitudeTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'c. Process Raw Altitude'
        self.description = '' 
        self.category = "b. QA/QC Tools"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Processing Project', 
                                 name = 'procProj', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Altitude Max Threshold', 
                                 name = 'altMax', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param3 = arcpy.Parameter(displayName = 'Rolling Stats Window Size', 
                                 name = 'window', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param4 = arcpy.Parameter(displayName = 'Altitude field', 
                                 name = 'inputAltfield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Line field', 
                                 name = 'inputLinefield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Time field', 
                                 name = 'inputTimefield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param7 = arcpy.Parameter(displayName = 'Output File Name (tool will append _altsmth)', 
                                 name = 'outputFileName', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        
        # Default values 
        param2.value = 20
        param3.value = 10
        param4.value = "Field2"
        param5.value = "Field3"
        param6.value = "Field4"

        param4.parameterDependencies = [param1.name]
        param5.parameterDependencies = [param1.name]
        param6.parameterDependencies = [param1.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        procProj=parameters[0].valueAsText        
        arcpy.AddMessage('------------------')
        arcpy.AddMessage('Processing Project: {0}'.format(procProj))
        arcpy.AddMessage('------------------')
        
        inputRawFile = parameters[1].valueAsText 
        arcpy.AddMessage('Raw File: {}'.format(inputRawFile))
        arcpy.AddMessage('------------------')

        altMax=(int(parameters[2].value))
        arcpy.AddMessage('Altitude Maximum Threshold: {}'.format(altMax))
        arcpy.AddMessage('------------------')

        window=(int(parameters[3].value))
        arcpy.AddMessage('Rolling Stats Window Size: {} points'.format(window))
        arcpy.AddMessage('------------------')

        inputAltfield=parameters[4].valueAsText

        inputLinefield=parameters[5].valueAsText

        inputTimefield=parameters[6].valueAsText

        altPlotPathway=(r'{0}\03_Processing_Figures\Altitude_Figures'.format(procProj))

        altSmthPathway=(r'{0}\02_Processing_Workspaces\AltSmth_temp'.format(procProj))

        outputWorkspace=(r'{0}\01_Data_Files\03_Smoothed_Files'.format(procProj))

        outputFileName=parameters[7].valueAsText 

        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Read raw file")
        arcpy.AddMessage('------------------')

        ##iterate a list of unique line names
        lineNameList=df['{}'.format(inputLinefield)].unique()
        arcpy.AddMessage(lineNameList)
        arcpy.AddMessage('------------------')
        arcpy.AddMessage("Total Number of Lines: {}".format(len(lineNameList)))
        arcpy.AddMessage('------------------')

        df['{}_dspk'.format(inputAltfield)]=df['{}'.format(inputAltfield)].where(df['{}'.format(inputAltfield)]< altMax)
        plt.rcParams.update({'figure.max_open_warning': 0})

        ## define a function to perform a rolling stats computation on single line of data
        def rollingStatsperLine(x): 
            arcpy.AddMessage("Processing Line: {}".format(lineNameList[x]))
            mask=df['{}'.format(inputLinefield)]==lineNameList[x]
            dfx=df[mask].copy()
            dfx['{}_intp'.format(inputAltfield)]=dfx['{}_dspk'.format(inputAltfield)].interpolate(method='linear')
            dfx['{}_smth'.format(inputAltfield)]=dfx['{}_intp'.format(inputAltfield)].rolling(window,center=True).mean()
            dfx.to_csv(r"{}\{}ln{}.txt".format(altSmthPathway,outputFileName,x),float_format='%.2f',index=False)
        ##plot raw channel, despike, and smoothed result
            fig = plt.figure()
            ax1=fig.add_subplot(311)
            plt.plot(dfx['{}'.format(inputTimefield)],dfx['{}'.format(inputAltfield)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            plt.ylabel("Altitude")
            minALT=df.round(1)['{0}'.format(inputAltfield)].min()
            maxALT=df.round(1)['{0}'.format(inputAltfield)].max()
            plt.ylim([minALT,maxALT])
            plt.legend(['Raw Alt'])
            ax1.axes.get_xaxis().set_visible(False)
       
            ax1 = fig.add_subplot(312)
            plt.plot(dfx['{}'.format(inputTimefield)],dfx['{}_dspk'.format(inputAltfield)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            plt.ylabel("Altitude")
            minALT=df.round(1)['{0}'.format(inputAltfield)].min()
            maxALT=df.round(1)['{0}'.format(inputAltfield)].max()
            plt.ylim([minALT,maxALT])
            plt.legend(['Dspk Alt'])
            ax1.axes.get_xaxis().set_visible(False)
        
            ax1 = fig.add_subplot(313)
            plt.plot(dfx['{}'.format(inputTimefield)],dfx['{}_smth'.format(inputAltfield)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            plt.ylabel("Altitude")
            minALT=df.round(1)['{0}'.format(inputAltfield)].min()
            maxALT=df.round(1)['{0}'.format(inputAltfield)].max()
            plt.ylim([minALT,maxALT])
            plt.legend(['Smth Alt'])
            ax1.axes.get_xaxis().set_visible(False)
            plt.savefig(r"{0}\{1}.jpg".format(altPlotPathway,lineNameList[x]))
            plt.close(fig)
            del fig
            return

        ## call the function and run as iterations of the line name list
        x=0
        while x <= len(lineNameList):
            rollingStatsperLine(x)
            x +=1
            if x == len(lineNameList):
                break

        ## read directory of CSV files generated for each df split by line, merge into a single df
        path =(r"{}".format(altSmthPathway))
        all_files = glob.glob(os.path.join(path,"{}*.txt".format(outputFileName)))
        df_from_each_file = (pd.read_csv(f,sep=',')for f in all_files)
        df_merged = pd.concat(df_from_each_file,ignore_index=True)
        df_final=df_merged.dropna()
        arcpy.AddMessage('------------------')
        arcpy.AddMessage("Merged Resutls")
        arcpy.AddMessage('------------------')

        ## export merged df as full csv file
        os.chdir(r"{}".format(outputWorkspace))
        df_final.to_csv("{}_altsmth.txt".format(outputFileName),index=False)

        for f in all_files:
            os.remove(f)

        arcpy.AddMessage("completed altitude smoothing")

class RawFileAssessmentTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'a. Assess Raw File'
        self.description = '' 
        self.category = "b. QA/QC Tools"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Processing Project', 
                                 name = 'procProj', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEastingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param3 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Total Field', 
                                 name = 'inputTotal Field', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Signal Quality Field', 
                                 name = 'inputSignalQuality', 
                                 datatype = 'Field', 
                                 parameterType = 'Optional', 
                                 direction = 'Input')

        param6 = arcpy.Parameter(displayName = 'Signal Quality Minimum Value', 
                                 name = 'SignalQualityMin', 
                                 datatype = 'String', 
                                 parameterType = 'Optional', 
                                 direction = 'Input')  

        param7 = arcpy.Parameter(displayName = 'Altitude Field', 
                                 name = 'inputAltField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param8 = arcpy.Parameter(displayName = 'Altitude Units', 
                                 name = 'inputAltUnits', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param9 = arcpy.Parameter(displayName = 'Survey Line Field', 
                                 name = 'inputSurveyLine', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        
        param10 = arcpy.Parameter(displayName = 'Time Field', 
                                 name = 'inputTime', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        # Default values 
        param2.value = "Field2"
        param3.value = "Field3"
        param4.value = "Feild4"
        param5.value = "Field5"
        param6.value = "450"
        param7.value = "Field6"
        param8.value = "Meters"
        param9.value = "Field7"
        param10.value = "Field8"

        param2.parameterDependencies = [param1.name]
        param3.parameterDependencies = [param1.name]
        param4.parameterDependencies = [param1.name]
        param5.parameterDependencies = [param1.name]
        param7.parameterDependencies = [param1.name]
        param9.parameterDependencies = [param1.name]
        param10.parameterDependencies = [param1.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9, param10] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        procProj=parameters[0].valueAsText
        arcpy.AddMessage('Processing Project: {0}'.format(procProj))
        arcpy.AddMessage('------------------')

        inputRawFile = parameters[1].valueAsText 
        arcpy.AddMessage('Raw File: {}'.format(inputRawFile))
        arcpy.AddMessage('------------------')

        Easting=parameters[2].valueAsText

        Northing=parameters[3].valueAsText

        inputTotalField=parameters[4].valueAsText

        inputSignalQuality=parameters[5].valueAsText

        SignalQualiyMin=parameters[6].valueAsText

        inputAltfield=parameters[7].valueAsText

        inputAltUnits=parameters[8].valueAsText

        inputSurveyLine=parameters[9].valueAsText

        inputTime=parameters[10].valueAsText

        qaqcPlotpathway=(r'{0}\03_Processing_Figures\QAQC_Summary_Figures'.format(procProj))

        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Converted Raw File to DataFrame")
        arcpy.AddMessage('------------------')

        ## iterate a list of unique line names
        lineNameList=df['{}'.format(inputSurveyLine)].unique()
        arcpy.AddMessage(lineNameList)
        arcpy.AddMessage("Total Number of Lines: {}".format(len(lineNameList)))
        arcpy.AddMessage("--------------------------------")
        plt.rcParams.update({'figure.max_open_warning': 0})

        def plotQAQCfigures(x):
            arcpy.AddMessage ("Assessing Line: {0}".format(lineNameList[x]))
            mask=df['{0}'.format(inputSurveyLine)]==lineNameList[x]
            dfx=df[mask].copy()

            fig = plt.figure()
            ax1 = fig.add_subplot(411)
            plt.plot(dfx['{0}'.format(inputTime)],dfx.round(1)['{0}'.format(Easting)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            ## plt.ylabel("Position", fontsize=8)
            plt.yticks(fontsize=8)
            plt.ticklabel_format(style='plain', useOffset=False, axis='y')
            plt.legend(['Easting'])
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(412)
            plt.plot(dfx['{0}'.format(inputTime)],dfx.round(1)['{0}'.format(Northing)], 'b-', linewidth=0.3)
            plt.xlabel("Time")
            ## plt.ylabel("Position", fontsize=8)
            plt.yticks(fontsize=8)
            plt.ticklabel_format(style='plain', useOffset=False, axis='y')
            plt.legend(['Northing'])
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(413)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}'.format(inputAltfield)], 'b-',linewidth=0.3)
            plt.xlabel("Time")
            ## plt.ylabel("Raw Altitude")
            minALT=df.round(1)['{0}'.format(inputAltfield)].min()
            maxALT=df.round(1)['{0}'.format(inputAltfield)].max()
            plt.ylim([minALT,maxALT])
            plt.legend(['Altitude'])
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(414)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}'.format(inputTotalField)], 'b-',linewidth=0.3)
            plt.xlabel("Time")
            ## plt.ylabel("Total Field (nT)")
            minTF=df.round(1)['{0}'.format(inputTotalField)].min()
            maxTF=df.round(1)['{0}'.format(inputTotalField)].max()
            plt.ylim([minTF,maxTF])
            plt.legend(['Total Field (nT)'])
            ax1.axes.get_xaxis().set_visible(False)

            plt.savefig(r"{0}\{1}.jpg".format(qaqcPlotpathway,lineNameList[x]), dpi=300)
            plt.close(fig)
            del fig
            return
        x=0
        while x <= len(lineNameList):
            plotQAQCfigures(x)
            x +=1
            if x == len(lineNameList):
                break
        
        arcpy.AddMessage("--------------------------------")
        arcpy.AddMessage("Computing Summary Statistics")
        arcpy.AddMessage("--------------------------------")
        ## summary stats for raw easting position data    
        maxEasting=df.round(2)['{0}'.format(Easting)].max()
        arcpy.AddMessage("Maximum Easting Value: {0}".format(maxEasting))
        meanEasting=df['{0}'.format(Easting)].mean().round(2)
        arcpy.AddMessage("Mean Easting Value: {0}".format(meanEasting))
        minEasting=df.round(2)['{0}'.format(Easting)].min()
        arcpy.AddMessage("Minimum Easting Value: {0}".format(minEasting))
        if minEasting == 0:
            arcpy.AddMessage("POSITIONING DROPOUTS DETECTED")
            arcpy.AddMessage("RECOMMEND CLEANING EASTING NAVIGATION CHANNEL")
        arcpy.AddMessage("--------------------------------")

        ## summary stats for raw northing position data
        maxNorthing=df.round(2)['{0}'.format(Northing)].max()
        arcpy.AddMessage("Maximum Northing Value: {0}".format(maxNorthing))
        meanNorthing=df['{0}'.format(Northing)].mean().round(2)
        arcpy.AddMessage("Mean Northing Value: {0}".format(meanNorthing))
        minNorthing=df.round(2)['{0}'.format(Northing)].min()
        arcpy.AddMessage("Minimum Northing Value: {0}".format(minNorthing))
        if minNorthing == 0:
            arcpy.AddMessage("POSITIONING DROPOUTS DETECTED")
            arcpy.AddMessage("RECOMMEND CLEANING NORTHING NAVIGATION CHANNEL")
        arcpy.AddMessage("--------------------------------")

        ## summary stats for raw altitude data
        maxAlt=df.round(2)['{0}'.format(inputAltfield)].max()
        arcpy.AddMessage("Maximum Raw Altitude: {0} {1}".format(maxAlt,inputAltUnits))
        meanAlt=df['{0}'.format(inputAltfield)].mean().round(2)
        arcpy.AddMessage("Mean Raw Altitude: {0} {1}".format(meanAlt,inputAltUnits))
        minAlt=df.round(2)['{0}'.format(inputAltfield)].min()
        arcpy.AddMessage("Minumum Raw Altitude: {0} {1}".format(minAlt,inputAltUnits))
        if minAlt == 0:
            arcpy.AddMessage("BOTTOM STRIKES DETECTED")
            arcpy.AddMessage("RECOMMEND CLEANING ALTITUDE CHANNEL")
        if maxAlt >= 15.0:
            arcpy.AddMessage("HIGH ALTITUDE VALUES and FLIERS DETECTED")
            arcpy.AddMessage("RECOMMEND CLEANING ALTITUDE CHANNEL")
        arcpy.AddMessage("--------------------------------")

        ## summary stats for total field data
        maxTF=df.round(2)['{0}'.format(inputTotalField)].max()
        arcpy.AddMessage("Maximum Total Field Value: {0} nT".format(maxTF))
        meanTF=df['{0}'.format(inputTotalField)].mean().round(2)
        arcpy.AddMessage("Mean Total Field Value: {0} nT".format(meanTF))
        minTF=df.round(2)['{0}'.format(inputTotalField)].min()
        arcpy.AddMessage("Minimum Total Field Value: {0} nT".format(minTF))
        arcpy.AddMessage("--------------------------------")

        ## summary stats for siqnal quality data
        maxSignalQ=df['{0}'.format(inputSignalQuality)].max()
        arcpy.AddMessage("Maximum Signal Quality Value: {0}".format(maxSignalQ))
        meanSignalQ=df['{0}'.format(inputSignalQuality)].mean().round(1)
        arcpy.AddMessage("Mean Signal Quality Value: {0}".format(meanSignalQ))
        minSignalQ=df['{0}'.format(inputSignalQuality)].min()
        arcpy.AddMessage("Minimum Signal Quality Value: {0}".format(minSignalQ))
        if minSignalQ <= int(SignalQualiyMin):
            arcpy.AddMessage("SIGNAL QUALITY DROPOUTS DETECTED")
            arcpy.AddMessage("RECOMMNED BLANKING TOTAL FIELD VALUES WITH LOW SIGNAL Q")
        arcpy.AddMessage("--------------------------------")

        arcpy.AddMessage("Completed Raw File QA/QC Check")

class ResidualFieldProcessingTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'b. Generate Residual Field'
        self.description = '' 
        self.category = "c. Magnetic Channel Processing Tools"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Processing Project', 
                                 name = 'procProj', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEastingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Total Field Channel', 
                                 name = 'inputTotalField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'First Filter Window Size', 
                                 name = 'firstFiltWind', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param6 = arcpy.Parameter(displayName = 'Second Filter Window Size', 
                                 name = 'secFiltWind', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Third Fitler Window Size', 
                                 name = 'thirdFiltWind', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input')                         

        param8 = arcpy.Parameter(displayName = 'Background Channel Smoothing Window Size', 
                                 name = 'bgSmooth', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param9 = arcpy.Parameter(displayName = 'Survey Line Field', 
                                 name = 'inputSurveyLine', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param10 = arcpy.Parameter(displayName = 'Time Field', 
                                 name = 'inputTimeField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param11 = arcpy.Parameter(displayName = 'Output File Name', 
                                 name = 'outfile', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        # Default values 
        param2.value = "Field2"
        param3.value = "Field3"
        param4.value = "Field4"
        param5.value = "5"
        param6.value = "11"
        param7.value = "51"
        param8.value = "20"
        param9.value = "Field5"
        param10.value = "Field6"

        param2.parameterDependencies = [param1.name]
        param3.parameterDependencies = [param1.name]
        param4.parameterDependencies = [param1.name]
        param9.parameterDependencies = [param1.name]
        param10.parameterDependencies = [param1.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9,param10, param11] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        procProj=parameters[0].valueAsText
        arcpy.AddMessage('Processing Project: {0}'.format(procProj))
        arcpy.AddMessage('------------------')

        inputRawFile = parameters[1].valueAsText 
        arcpy.AddMessage('Raw File: {}'.format(inputRawFile))
        arcpy.AddMessage('------------------')

        inputEasting=parameters[2].value

        inputNorthing=parameters[3].value
        
        inputTotalField=parameters[4].valueAsText

        firstFiltWind=parameters[5].valueAsText

        secFiltWind=parameters[6].valueAsText

        thirdFiltWind=parameters[7].valueAsText

        bgSmooth=int(parameters[8].valueAsText)

        inputSurveyLine=parameters[9].valueAsText

        inputTime=parameters[10].valueAsText

        residualPlotpathway=(r'{0}\03_Processing_Figures\Residual_Figures'.format(procProj))

        resProcWorkspace=(r'{0}\02_Processing_Workspaces\Residual_temp'.format(procProj))

        outfilepath=(r'{0}\01_Data_Files\04_Processed_Files'.format(procProj))

        outfilename=parameters[11].valueAsText

        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Converted Raw File to DataFrame")
        arcpy.AddMessage('------------------')

        ## iterate a list of unique line names
        lineNameList=df['{}'.format(inputSurveyLine)].unique()
        arcpy.AddMessage(lineNameList)
        arcpy.AddMessage('------------------')
        arcpy.AddMessage("Total Number of Lines: {}".format(len(lineNameList)))
        arcpy.AddMessage('------------------')
        plt.rcParams.update({'figure.max_open_warning': 0})

        def background_filter(x):
            arcpy.AddMessage ("Processing Line: {0}".format(lineNameList[x]))
            mask=df['{0}'.format(inputSurveyLine)]==lineNameList[x]
            dfx=df[mask].copy()

            ##  calculate a distance field in the dataframe
            # dfx['easting1']=dfx['{0}'.format(inputEasting)]
            # dfx['easting2']=dfx['{0}'.format(inputEasting)].shift(-1)
            # dfx['northing1']=dfx['{0}'.format(inputNorthing)]
            # dfx['northing2']=dfx['{0}'.format(inputNorthing)].shift(-1)
            # dfx['dist']=numpy.sqrt((dfx['easting1']-dfx['easting2'])**2 + (dfx['northing1']-dfx['northing2'])**2 )
            # del dfx['easting1']
            # del dfx['easting2']
            # del dfx['northing1']
            # del dfx['northing2']
 
        ##  Using sequential Savitzky-Golay filters with changing window sizes

            totalField=dfx['{0}'.format(inputTotalField)]
            Win1=int(firstFiltWind)
            polyorder=1
            deriv=0
            dfx['nT_filt1']=(signal.savgol_filter(totalField,Win1,polyorder,deriv,mode='mirror'))

            firstFilt=dfx['nT_filt1']
            Win2=int(secFiltWind)
            polyorder=1
            deriv=0
            dfx['nT_filt2']=(signal.savgol_filter(firstFilt,Win2,polyorder,deriv,mode='mirror'))

            secFilt=dfx['nT_filt2']
            Win3=int(thirdFiltWind)
            polyorder=1
            deriv=0
            dfx['nT_filt3']=(signal.savgol_filter(secFilt,Win3,polyorder,deriv,mode='mirror'))
            thirdFilt=dfx['nT_filt3']
            dfx['nT_background']=(signal.lfilter(thirdFilt,thirdFilt,thirdFilt)).round(2)

        ##  smooth the filtered background resuls
            ##dfx.round(2)['nT_background']=(dfx['nT_background'].rolling(bgSmooth,center=True).mean())    

        ##  caluclate residual field by subtracting background from total field    
            dfx['nT_res']=dfx['nT_background']-dfx['{0}'.format(inputTotalField)]
            
            dfx.to_csv(r"{0}\res_grid_{1}.txt".format(resProcWorkspace,x),float_format='%.2f',index=False)
            
            fig = plt.figure()
            ax1 = fig.add_subplot(211)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}'.format(inputTotalField)], 'b-', linewidth=0.3)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['nT_background'], 'r-',linewidth=0.4)
            plt.ylabel("Total Field (nT)")
            minTF=dfx.round(2)['{0}'.format(inputTotalField)].min()-50
            maxTF=dfx.round(2)['{0}'.format(inputTotalField)].max()+50
            plt.ylim([minTF,maxTF])
            plt.legend(['Original','Background'])
            plt.title("Total Field Vs. Filtered Signal")
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(212)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['nT_res'], 'b-', linewidth=0.5)
            plt.ylabel("Residual Field")
            plt.ylim([-15,15])
            plt.xlabel("Time")
            plt.legend(['Residuals'])
            ax1.axes.get_xaxis().set_visible(False)
            plt.savefig(r"{0}\{1}.jpg".format(residualPlotpathway,lineNameList[x]), dpi=300)
            plt.close(fig)
            del fig
            return
        x=0
        while x <= len(lineNameList):
            background_filter(x)
            x +=1
            if x == len(lineNameList):
                break

        path =(r"{0}".format(resProcWorkspace))
        all_files = glob.glob(os.path.join(path,"res_grid*.txt"))
        df_from_each_file = (pd.read_csv(f,sep=',')for f in all_files)
        df_merged = pd.concat(df_from_each_file,ignore_index=True)
        arcpy.AddMessage(df_merged.columns)
        arcpy.AddMessage(df_merged.size)
        df_final=df_merged.dropna()
        arcpy.AddMessage(df_final.columns)

        ## export merged df as full csv file
        os.chdir(r"{0}".format(outfilepath))
        df_final.to_csv("{0}.txt".format(outfilename),index=False)

        for f in all_files:
            os.remove(f)
        
        arcpy.AddMessage('------------------')
        arcpy.AddMessage("Completed Residual Field Creation")
        arcpy.AddMessage('------------------')

class SpatialGradientProcessingTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'c. Generate Spatial Gradient Field'
        self.description = '' 
        self.category = "c. Magnetic Channel Processing Tools"
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Processing Project', 
                                 name = 'procProj', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEastingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')
        
        param4 = arcpy.Parameter(displayName = 'Total Field Channel', 
                                 name = 'inputTotalField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Survey Line Field', 
                                 name = 'inputSurveyLine', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Time Field', 
                                 name = 'inputTimeLine', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Output File Name', 
                                 name = 'outfile', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input')                         

        # Default values 
        param2.value = "Field1"
        param3.value = "Field2"
        param4.value = "Field3"
        param5.value = "Field4"
        param6.value = "Field5"

        param2.parameterDependencies = [param1.name]
        param3.parameterDependencies = [param1.name]
        param4.parameterDependencies = [param1.name]
        param5.parameterDependencies = [param1.name]
        param6.parameterDependencies = [param1.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        procProj=parameters[0].valueAsText
        arcpy.AddMessage('Processing Project: {0}'.format(procProj))
        arcpy.AddMessage('------------------')

        inputRawFile = parameters[1].valueAsText 
        arcpy.AddMessage('Raw File: {}'.format(inputRawFile))
        arcpy.AddMessage('------------------')

        inputEasting=parameters[2].valueAsText

        inputNorthing=parameters[3].valueAsText

        inputTotalField=parameters[4].valueAsText

        inputSurveyLine=parameters[5].valueAsText

        inputTime=parameters[6].valueAsText

        spatialgradPlotpathway=(r'{0}\03_Processing_Figures\SpatialGradient_Figures'.format(procProj))

        spatialgradProcWorkspace=(r'{0}\02_Processing_Workspaces\Gradient_temp'.format(procProj))

        outfilepath=(r'{0}\01_Data_Files\04_Processed_Files'.format(procProj))

        outfilename=parameters[7].valueAsText

        
        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Converted Raw File to DataFrame")

        ## iterate a list of unique line names
        lineNameList=df['{}'.format(inputSurveyLine)].unique()
        arcpy.AddMessage(lineNameList)
        arcpy.AddMessage("Total Number of Lines: {}".format(len(lineNameList)))
        arcpy.AddMessage("--------------------------------")
        plt.rcParams.update({'figure.max_open_warning': 0})

        def spatialGradient_filter(x):
            arcpy.AddMessage ("Processing Line: {0}".format(lineNameList[x]))
            mask=df['{0}'.format(inputSurveyLine)]==lineNameList[x]
            dfx=df[mask].copy()
 
        ##  calculate a distance field in the dataframe
            dfx['easting1']=dfx['{0}'.format(inputEasting)]
            dfx['easting2']=dfx['{0}'.format(inputEasting)].shift(-1)
            dfx['northing1']=dfx['{0}'.format(inputNorthing)]
            dfx['northing2']=dfx['{0}'.format(inputNorthing)].shift(-1)
            dfx['dist']=numpy.sqrt((dfx['easting1']-dfx['easting2'])**2 + (dfx['northing1']-dfx['northing2'])**2 )
            del dfx['easting1']
            del dfx['easting2']
            del dfx['northing1']
            del dfx['northing2']

        ##  calcuate the difference between sequential total field readings
            dfx['totalField_Diff']=dfx['{0}'.format(inputTotalField)].diff()
        
        ## calcuate the spatial gradient 
            dfx['spatialGradient']=dfx['totalField_Diff']/dfx['dist']
            dfx['absGradient']=dfx['spatialGradient'].abs()

            dfx.to_csv(r"{0}\spat_grad_{1}.txt".format(spatialgradProcWorkspace,x),float_format='%.2f',index=False)
            
            fig = plt.figure()
            ax1 = fig.add_subplot(211)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['{0}'.format(inputTotalField)], 'b-', linewidth=0.3)
            plt.ylabel("Total Field (nT)")
            minTF=dfx.round(2)['{0}'.format(inputTotalField)].min()-50
            maxTF=dfx.round(2)['{0}'.format(inputTotalField)].max()+50
            plt.ylim([minTF,maxTF])
            plt.title("Total Field")
            ax1.axes.get_xaxis().set_visible(False)

            ax1 = fig.add_subplot(212)
            plt.plot(dfx['{0}'.format(inputTime)],dfx['spatialGradient'], 'b-', linewidth=0.3)
            plt.ylabel("Spatial Gradient")
            plt.ylim([-15,15])
            plt.xlabel("Time")
            plt.legend(['Spatial Gradient Result'])
            ax1.axes.get_xaxis().set_visible(False)
            plt.savefig(r"{0}\{1}.jpg".format(spatialgradPlotpathway,lineNameList[x]), dpi=300)
            plt.close(fig)
            del fig
            return
        x=0
        while x <= len(lineNameList):
            spatialGradient_filter(x)
            x +=1
            if x == len(lineNameList):
                break

        path =(r"{0}".format(spatialgradProcWorkspace))
        all_files = glob.glob(os.path.join(path,"spat_grad*.txt"))
        df_from_each_file = (pd.read_csv(f,sep=',')for f in all_files)
        df_merged = pd.concat(df_from_each_file,ignore_index=True)
        arcpy.AddMessage(df_merged.columns)
        arcpy.AddMessage(df_merged.size)
        df_final=df_merged.dropna()
        arcpy.AddMessage(df_final.columns)

        ## export merged df as full csv file
        os.chdir(r"{0}".format(outfilepath))
        df_final.to_csv("{0}.txt".format(outfilename),index=False)

        for f in all_files:
            os.remove(f)
        
        arcpy.AddMessage("Completed Spatial Gradient Field Creation")

class CreateNewProcesseingProjectTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Create New Processing Project Archive'
        self.description = '' 
        self.category = 'a. Project Management'
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'New Project Name', 
                                 name = 'newProjectName', 
                                 datatype = 'String', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'New Project Workspace', 
                                 name = 'projWorkspace', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        # Default values 
        param0.value = "YYYYNN_Mag_Project"

        return [param0, param1] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        projName= parameters[0].valueAsText 
        arcpy.AddMessage('New Project Name: {}'.format(projName))
        arcpy.AddMessage('-----------------------------')
        arcpy.AddMessage('-----------------------------')

        projWorksapce=parameters[1].valueAsText

        arcpy.AddMessage("Starting New Project Directory")
        arcpy.AddMessage('-----------------------------')
        arcpy.AddMessage('-----------------------------')

        path=(r'{0}\{1}\01_Data_Files'.format(projWorksapce,projName))
        os.makedirs(path)
        path2=(r'{0}\{1}\01_Data_Files\01_Raw_Files'.format(projWorksapce,projName))
        path3=(r'{0}\{1}\01_Data_Files\02_ReFormatted_Files'.format(projWorksapce,projName))
        path4=(r'{0}\{1}\01_Data_Files\03_Smoothed_Files'.format(projWorksapce,projName))
        path5=(r'{0}\{1}\01_Data_Files\04_Processed_Files'.format(projWorksapce,projName))
        os.mkdir(path2)
        os.mkdir(path3)
        os.mkdir(path4)
        os.mkdir(path5)
        path6=(r'{0}\{1}\02_Processing_Workspaces'.format(projWorksapce,projName))
        os.makedirs(path6)
        path7=(r'{0}\{1}\02_Processing_Workspaces\AltSmth_temp'.format(projWorksapce,projName))
        path8=(r'{0}\{1}\02_Processing_Workspaces\Gradient_temp'.format(projWorksapce,projName))
        path9=(r'{0}\{1}\02_Processing_Workspaces\NavSmth_temp'.format(projWorksapce,projName))
        path10=(r'{0}\{1}\02_Processing_Workspaces\Residual_temp'.format(projWorksapce,projName))
        os.mkdir(path7)
        os.mkdir(path8)
        os.mkdir(path9)
        os.mkdir(path10)

        path11=(r'{0}\{1}\03_Processing_Figures'.format(projWorksapce,projName))
        os.makedirs(path11)
        path12=(r'{0}\{1}\03_Processing_Figures\Altitude_Figures'.format(projWorksapce,projName))
        path13=(r'{0}\{1}\03_Processing_Figures\Navigation_Figures'.format(projWorksapce,projName))
        path14=(r'{0}\{1}\03_Processing_Figures\QAQC_Summary_Figures'.format(projWorksapce,projName))
        path15=(r'{0}\{1}\03_Processing_Figures\Residual_Figures'.format(projWorksapce,projName))
        path16=(r'{0}\{1}\03_Processing_Figures\SpatialGradient_Figures'.format(projWorksapce,projName))
        os.mkdir(path12)
        os.mkdir(path13)
        os.mkdir(path14)
        os.mkdir(path15)
        os.mkdir(path16)
        path17=(r'{0}\{1}\04_GIS_Workspace'.format(projWorksapce,projName))
        os.makedirs(path17)

        
        arcpy.AddMessage("Created New Project Direcotry")
        arcpy.AddMessage('-----------------------------')
        arcpy.AddMessage('-----------------------------')

class ConvertRawFileSchema(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Reformat Single Mag File'
        self.description = '' 
        self.category = 'a. Project Management'
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Processing Project', 
                                 name = 'procProj', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEastingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')
        
        param4 = arcpy.Parameter(displayName = 'Total Field Channel', 
                                 name = 'inputTotalField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Siqnal Quality Channel', 
                                 name = 'inputSigQField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param6 = arcpy.Parameter(displayName = 'Sensor Altitude Channel', 
                                name = 'inputAltField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param7 = arcpy.Parameter(displayName = 'Survey Line Field', 
                                 name = 'inputSurveyLine', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param8 = arcpy.Parameter(displayName = 'Date Field', 
                                 name = 'inputDateField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param9 = arcpy.Parameter(displayName = 'Time Field', 
                                 name = 'inputTimeField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')  

        param10 = arcpy.Parameter(displayName = 'Output File Name', 
                                 name = 'outfile', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input')                         

        # Default values
        param0.value = "YYYYNN_Mag_Project" 
        param2.value = "Field1"
        param3.value = "Field2"
        param4.value = "Field3"
        param5.value = "Field4"
        param6.value = "Field5"
        param7.value = "Field6"
        param8.value = "Field7"
        param9.value = "Field8"

        param2.parameterDependencies = [param1.name]
        param3.parameterDependencies = [param1.name]
        param4.parameterDependencies = [param1.name]
        param5.parameterDependencies = [param1.name]
        param6.parameterDependencies = [param1.name]
        param7.parameterDependencies = [param1.name]
        param8.parameterDependencies = [param1.name]
        param9.parameterDependencies = [param1.name]
        
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9, param10] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        procProj= parameters[0].valueAsText 
        arcpy.AddMessage('Processing Project Name: {}'.format(procProj))
        arcpy.AddMessage('-----------------------------')

        inputRawFile=parameters[1].valueAsText
        arcpy.AddMessage('Raw File Name: {}'.format(inputRawFile))
        arcpy.AddMessage('-----------------------------')
        
        inputEasting=parameters[2].valueAsText

        inputNorthing=parameters[3].valueAsText

        inputTotalField=parameters[4].valueAsText

        inputSiqnalQ=parameters[5].valueAsText

        inputAltitude=parameters[6].valueAsText

        inputSurveyLine=parameters[7].valueAsText

        inputDate=parameters[8].valueAsText

        inputTime=parameters[9].valueAsText

        outfilepath=(r'{0}\01_Data_Files\02_ReFormatted_Files'.format(procProj))

        outfilename=parameters[10].valueAsText

        ## convert CSV file to data frame
        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Converted Raw File to DataFrame")
        arcpy.AddMessage('-----------------------------')
        inputHeader=list(df.columns)
        arcpy.AddMessage('Raw File Headers: {0}'.format(inputHeader))

        ## create new headers for colmuns in data frame
        df.rename(columns={inputEasting: 'easting',
                                inputNorthing: 'northing',
                                inputTotalField: 'nT_raw',
                                inputSiqnalQ: 'signalQ',
                                inputAltitude: 'altitude',
                                inputSurveyLine: 'line_file',
                                inputDate: 'date',
                                inputTime: 'time'}, inplace=True)

        ## below items repleaced by above syntax to fix error message
        ## df['easting']=df[{inputEasting}]
        ## df['northing']=df[{inputNorthing}]
        ## df['nT_raw']=df[{inputTotalField}]
        ## df['signalQ']=df[{inputSiqnalQ}]
        ## df['altitude']=df[{inputAltitude}]
        ## df['line_file']=df[{inputSurveyLine}]
        ## df['date']=df[{inputDate}]
        ## df['time']=df[{inputTime}]
        arcpy.AddMessage('-----------------------------')
        arcpy.AddMessage("New Dataframe Headers Created")
        arcpy.AddMessage('-----------------------------')
        newHeader=list(df.columns)
        arcpy.AddMessage('New Dataframe Headers: {0}'.format(newHeader))
        arcpy.AddMessage('-----------------------------')
        df_new=df[['date','time','easting','northing','nT_raw','signalQ','altitude','line_file']]
        arcpy.AddMessage("Fields Re-Ordered")
        arcpy.AddMessage('-----------------------------')

        os.chdir(r"{0}".format(outfilepath))
        df_new.to_csv("{0}.txt".format(outfilename),index=False)
        
        arcpy.AddMessage("Converted Single Mag File to Processing Schema")
        arcpy.AddMessage('-----------------------------')
        arcpy.AddMessage('-----------------------------')

class ConvertGradiometerFile(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Reformat Gradiometer File'
        self.description = '' 
        self.category = 'a. Project Management'
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Processing Project', 
                                 name = 'procProj', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Input Gradiometer File', 
                                 name = 'inputGradFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Port Sensor Easting Field', 
                                 name = 'inputPSEastingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Port Sensor Northing Field', 
                                 name = 'inputPSNorthingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')
        
        param4 = arcpy.Parameter(displayName = 'Port Sensor Total Field Channel', 
                                 name = 'inputPSTotalField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Port Sensor Siqnal Quality Channel', 
                                 name = 'inputPSSigQField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param6 = arcpy.Parameter(displayName = 'Port Sensor Sensor Altitude Channel', 
                                name = 'inputPSAltField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param7 = arcpy.Parameter(displayName = 'Starboard Sensor Easting Field', 
                                 name = 'inputSSEastingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param8 = arcpy.Parameter(displayName = 'Starboard Sensor Northing Field', 
                                 name = 'inputSSNorthingField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')
        
        param9 = arcpy.Parameter(displayName = 'Starboard Sensor Total Field Channel', 
                                 name = 'inputSSTotalField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param10 = arcpy.Parameter(displayName = 'Starboard Sensor Siqnal Quality Channel', 
                                 name = 'inputSSSigQField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param11 = arcpy.Parameter(displayName = 'Starboard Sensor Sensor Altitude Channel', 
                                name = 'inputSSAltField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param12 = arcpy.Parameter(displayName = 'Survey Line Field', 
                                 name = 'inputSurveyLine', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param13 = arcpy.Parameter(displayName = 'Date Field', 
                                 name = 'inputDateField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param14 = arcpy.Parameter(displayName = 'Time Field', 
                                 name = 'inputTimeField', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')  


        param15 = arcpy.Parameter(displayName = 'Output File Name', 
                                 name = 'outfile', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input')                         

        # Default values 
        param0.value = "YYYYNN_Mag_Project"
        param2.value = "Field1"
        param3.value = "Field2"
        param4.value = "Field3"
        param5.value = "Field4"
        param6.value = "Field5"
        param7.value = "Field6"
        param8.value = "Field7"
        param9.value = "Field8"
        param10.value = "Field9"
        param11.value = "Field10"
        param12.value = "Field11"
        param13.value = "Field12"
        param14.value = "Field13"

        param2.parameterDependencies = [param1.name]
        param3.parameterDependencies = [param1.name]
        param4.parameterDependencies = [param1.name]
        param5.parameterDependencies = [param1.name]
        param6.parameterDependencies = [param1.name]
        param7.parameterDependencies = [param1.name]
        param8.parameterDependencies = [param1.name]
        param9.parameterDependencies = [param1.name]
        param10.parameterDependencies = [param1.name]
        param11.parameterDependencies = [param1.name]
        param12.parameterDependencies = [param1.name]
        param13.parameterDependencies = [param1.name]
        param14.parameterDependencies = [param1.name]
        
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11, param12, param13, param14, param15] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        procProj= parameters[0].valueAsText 
        arcpy.AddMessage('Processing Project Name: {}'.format(procProj))
        arcpy.AddMessage('-----------------------------')

        inputRawFile=parameters[1].valueAsText
        arcpy.AddMessage('Raw File Name: {}'.format(inputRawFile))
        arcpy.AddMessage('-----------------------------')
        
        inputPSEasting=parameters[2].valueAsText

        inputPSNorthing=parameters[3].valueAsText

        inputPSTotalField=parameters[4].valueAsText

        inputPSSiqnalQ=parameters[5].valueAsText

        inputPSAltitude=parameters[6].valueAsText

        inputSSEasting=parameters[7].valueAsText

        inputSSNorthing=parameters[8].valueAsText

        inputSSTotalField=parameters[9].valueAsText

        inputSSSiqnalQ=parameters[10].valueAsText

        inputSSAltitude=parameters[11].valueAsText

        inputSurveyLine=parameters[12].valueAsText

        inputDate=parameters[13].valueAsText

        inputTime=parameters[14].valueAsText

        outfilepath=(r'{0}\01_Data_Files\02_ReFormatted_Files'.format(procProj))

        outfilename=parameters[15].valueAsText

        ## convert CSV file to data frame, generated data frame for port sensor values
        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df_portsensor=pd.DataFrame(inputTable)
        arcpy.AddMessage("Converted Raw File to DataFrame")
        arcpy.AddMessage('-----------------------------')
        inputHeader=list(df_portsensor.columns)
        arcpy.AddMessage('Raw File Headers: {0}'.format(inputHeader))
        arcpy.AddMessage('-----------------------------')

        ## create new headers for colmuns in data frame

        df_portsensor.rename(columns={inputPSEasting: 'easting',
                                inputPSNorthing: 'northing',
                                inputPSTotalField: 'nT_raw',
                                inputPSSiqnalQ: 'signalQ',
                                inputPSAltitude: 'altitude',
                                inputSurveyLine: 'line_file',
                                inputDate: 'date',
                                inputTime: 'time'}, inplace=True)
        
        df_portsensor['sensor']="_PORT"
        df_portsensor['line_file']=df_portsensor[['line_file','sensor']].apply(lambda x:''.join(x),axis=1)

        ## below code replaced with above
        ## df_portsensor['easting']=df_portsensor[{inputPSEasting}]
        ## df_portsensor['northing']=df_portsensor[{inputPSNorthing}]
        ## df_portsensor['nT_raw']=df_portsensor[{inputPSTotalField}]
        ## df_portsensor['signalQ']=df_portsensor[{inputPSSiqnalQ}]
        ## df_portsensor['altitude']=df_portsensor[{inputPSAltitude}]
        ## df_portsensor['line_name']=df_portsensor[{inputSurveyLine}]
        ## df_portsensor['sensor']="_PORT"
        ## df_portsensor['line_file']=df_portsensor[['line_name','sensor']].apply(lambda x:''.join(x),axis=1)
        ## df_portsensor['date']=df_portsensor[{inputDate}]
        ## df_portsensor['time']=df_portsensor[{inputTime}]
        arcpy.AddMessage("Port Sensor Data Frame Created")
        arcpy.AddMessage('-----------------------------')
        inputHeader=list(df_portsensor.columns)
        arcpy.AddMessage('Raw File Headers: {0}'.format(inputHeader))
        arcpy.AddMessage('-----------------------------')
        df_portsensor_new=df_portsensor[['date','time','easting','northing','nT_raw','signalQ','altitude','line_file']]
        arcpy.AddMessage("Port Sensor Fields Re-Ordered")
        arcpy.AddMessage('-----------------------------')
        inputHeader=list(df_portsensor_new.columns)
        arcpy.AddMessage('Port Sensor Fields: {0}'.format(inputHeader))
        arcpy.AddMessage('-----------------------------')
        
        ## convert CSV file to data frame, generated data frame for starboard sensor values
        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df_stbdsensor=pd.DataFrame(inputTable)
        arcpy.AddMessage("Converted Raw File to DataFrame")
        arcpy.AddMessage('-----------------------------')
        inputHeader=list(df_stbdsensor.columns)
        arcpy.AddMessage('Raw File Headers: {0}'.format(inputHeader))
        arcpy.AddMessage('-----------------------------')
        
        ## create new headers for colmuns in data frame
        df_stbdsensor.rename(columns={inputSSEasting: 'easting',
                                inputSSNorthing: 'northing',
                                inputSSTotalField: 'nT_raw',
                                inputSSSiqnalQ: 'signalQ',
                                inputSSAltitude: 'altitude',
                                inputSurveyLine: 'line_file',
                                inputDate: 'date',
                                inputTime: 'time'}, inplace=True)
        
        df_stbdsensor['sensor']="_STBD"
        df_stbdsensor['line_file']=df_stbdsensor[['line_file','sensor']].apply(lambda x:''.join(x),axis=1)
        
        ## below code replaced with above
        ## df_stbdsensor['easting']=df_stbdsensor[{inputSSEasting}]
        ## df_stbdsensor['northing']=df_stbdsensor[{inputSSNorthing}]
        ## df_stbdsensor['nT_raw']=df_stbdsensor[{inputSSTotalField}]
        ## df_stbdsensor['signalQ']=df_stbdsensor[{inputSSSiqnalQ}]
        ## df_stbdsensor['altitude']=df_stbdsensor[{inputSSAltitude}]
        ## df_stbdsensor['line_name']=df_stbdsensor[{inputSurveyLine}]
        ## df_stbdsensor['sensor']="_STBD"
        ## df_stbdsensor['line_file']=df_stbdsensor[['line_name','sensor']].apply(lambda x:''.join(x),axis=1)
        ## df_stbdsensor['date']=df_stbdsensor[{inputDate}]
        ## df_stbdsensor['time']=df_stbdsensor[{inputTime}]
        arcpy.AddMessage("Starboard Sensor Data Frame Created")
        arcpy.AddMessage('-----------------------------')
        inputHeader=list(df_stbdsensor.columns)
        arcpy.AddMessage('Raw File Headers: {0}'.format(inputHeader))
        arcpy.AddMessage('-----------------------------')
        df_stbdsensor_new=df_stbdsensor[['date','time','easting','northing','nT_raw','signalQ','altitude','line_file']]
        arcpy.AddMessage("Starboard Sensor Fields Re-Ordered")
        arcpy.AddMessage('-----------------------------')
        inputHeader=list(df_stbdsensor_new.columns)
        arcpy.AddMessage('Starboard Sensor Sensor Fields: {0}'.format(inputHeader))

        ##concatenate data frames to
        df_gradmerge=pd.concat([df_portsensor_new, df_stbdsensor_new])
        arcpy.AddMessage('-----------------------------')
        arcpy.AddMessage('Sensor Data Frames Merged')
        arcpy.AddMessage('-----------------------------')

        os.chdir(r"{0}".format(outfilepath))
        df_gradmerge.to_csv("{0}.txt".format(outfilename),index=False)
        
        arcpy.AddMessage("Converted Gradiometer File to Processing Schema")
        arcpy.AddMessage('-----------------------------')
        arcpy.AddMessage('-----------------------------')